# -*- coding: utf-8 -*-
#
# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Access and handling of topographic data
"""
import numpy as np
import os

from geonum.exceptions import InvalidTopoMode, TopoAccessError
from geonum.topoaccessbase import SRTMAccess, Etopo1Access

class TopoDataAccess(object):
    """Factory class for accessing topographic data

    This is a high-level factory class which handles the access of topography
    data. Registered topographic datasets can be found in :attr:`REGISTERED`

    Default access mode is SRTM (topographic dataset from NASA
    `Shuttle Radar Topography Mission <https://www2.jpl.nasa.gov/srtm/>`__).

    Example
    -------

    >>> acc = TopoDataAccess(mode='srtm')
    >>> topo_data = acc.get_data(lat0=5, lon0=30, lat1=15, lon1=40)

    Note
    ----

    For developers: the registered access  classes (such as
    :class:`SRTMAccess`) should be based on (or follow the API of) template
    base class :class:`TopoAccessBase`.

    Attributes
    ----------
    mode : str
        current access mode (string specifying which topographic dataset is
        supposed to be used for access).
    local_path : str
        local path to etopo data (only relevant for etopo1 mode)

    Parameters
    ----------
    mode : str
        one of the supported data access types (cf. keys of :attr:`REGISTERED`)
    local_path : str
        local path to etopo data (only relevant for etopo1 mode)
    """
    #: supported access modes (topographic datasets)
    REGISTERED = dict(srtm = SRTMAccess,
                      etopo1 = Etopo1Access)
    DEFAULT_MODE = 'srtm'
    def __init__(self, mode=None, local_path=None):
        if mode is None:
            mode = self.DEFAULT_MODE
        #mode and data access variables
        if not mode in self.REGISTERED:
            raise InvalidTopoMode(mode)
        self.mode = mode

        self.local_path = local_path

    @property
    def modes(self):
        """List of supported topographic datasets"""
        return list(self.REGISTERED.keys())

    @property
    def supported(self):
        """List of supported datasets (wrapper for :attr:`modes`)"""
        return self.modes

    def __deepcopy__(self, memo):
        return TopoDataAccess(self.mode, self.local_path)

    def get_data(self, lat0, lon0, lat1=None, lon1=None,
                 mode=None, local_path=None, check_allnan=True,
                 **access_opts):
        """Retrieve data from topography file

        Parameters
        ----------
        lat0 : float
            start longitude for data extraction
        lon0 : float
            start latitude for data extraction
        lat1 : float
            stop longitude for data extraction (default: None). If None only
            data around lon0, lat0 will be extracted.
        lon1 : float
            stop latitude for data extraction (default: None).
            If None only data around lon0, lat0 will be extracted
        mode : str, optional
            mode specifying the topographic dataset that is supposed to be used
        local_path : str, optional
            local path where topography data is stored (can be dictionary or
            filepath, is passed to corresponding access class and handled
            as implemented there)
        check_allnan : bool
            if True and all retrieved values are NaN, then an error is thrown
        **access_opts
            additional access options that may be specific for the mode
            specified (e.g. search_database in case of etopo1)

        Returns
        -------
        TopoData
            object containing the data

        Raises
        ------
        TopoAccessError
            if access fails

        """
        if mode is not None:
            if not mode in self.REGISTERED:
                raise InvalidTopoMode(mode)
            self.mode = mode
        if local_path is not None and os.path.exists(local_path):
            self.local_path=local_path
        access = self.REGISTERED[self.mode](local_path=self.local_path,
                                             **access_opts)

        data = access.get_data(lat0, lon0, lat1, lon1)
        if check_allnan and np.isnan(data.data).all():
            raise TopoAccessError('Failed to retrieve TopoData ({}): '
                                  'all values are NaN'.format(self.mode))
        return data


import matplotlib.figure
import numpy as np
import pytest

from geonum import LineOnGrid


@pytest.mark.parametrize('line,result', [
    (LineOnGrid(0, 0, 0, 1), (1, 0)),
    (LineOnGrid(0, 0, 1, 0), (0, 1)),
    (LineOnGrid(0, 0, 1, 1), (1 / np.sqrt(2), 1 / np.sqrt(2))),
])
def test_LineOnGrid__det_normal_vec(line, result):
    vec = line._det_normal_vec()
    assert isinstance(vec, tuple)
    np.testing.assert_allclose(vec, result, atol=1e-4)


@pytest.mark.parametrize('line,result', [
    (LineOnGrid(0, 0, 0, 1), (0, 1)),
    (LineOnGrid(0, 0, 1, 0), (1, 0)),
    (LineOnGrid(2, 12, 44, 14), (42, 2)),
])
def test_LineOnGrid__delx_dely(line, result):
    vec = line._delx_dely()
    assert isinstance(vec, tuple)
    np.testing.assert_allclose(vec, result, atol=1e-4)


def test_LineOnGrid___str__():
    line = LineOnGrid(1, 1, 2, 2, name='bla')
    st = str(line)
    assert st == 'LineOnGrid bla. Start (x,y): [1, 1]. Stop (x,y): [2, 2]'


def test_LineOnGrid_plot():
    val = LineOnGrid(1, 1, 4, 6).plot(np.ones((20, 30)))
    assert isinstance(val, matplotlib.figure.Figure)


def _check_requirements():
    BASEMAP_AVAILABLE = True
    CV2_AVAILABLE = True
    NETCDF_AVAILABLE = True
    try:
        from mpl_toolkits.basemap import Basemap
    except ModuleNotFoundError:
        BASEMAP_AVAILABLE = False

    try:
        from cv2 import pyrUp
    except ModuleNotFoundError: # pragma: no cover
        CV2_AVAILABLE = False
    try:
        from netCDF4 import Dataset
    except ModuleNotFoundError: # pragma: no cover
        NETCDF_AVAILABLE = False

    return (BASEMAP_AVAILABLE,
            CV2_AVAILABLE,
            NETCDF_AVAILABLE)


def _init_local_topodir():
    try:
        import os
        home = os.path.expanduser('~')
        LOCAL_TOPO_DIR = os.path.join(home, '.geonum')
        if not os.path.exists(LOCAL_TOPO_DIR): # pragma: no cover
            os.mkdir(LOCAL_TOPO_DIR)
        TOPO_INFO_FILE = os.path.join(LOCAL_TOPO_DIR,  "LOCAL_TOPO_PATHS")
        if not os.path.exists(TOPO_INFO_FILE): # pragma: no cover
            with open(TOPO_INFO_FILE,'w') as f:
                f.write(f'{LOCAL_TOPO_DIR}\n')
    except Exception: # pragma: no cover
        print('Failed to create local topo directory for geonum')
        LOCAL_TOPO_DIR, TOPO_INFO_FILE = None, None
    return (LOCAL_TOPO_DIR, TOPO_INFO_FILE)


def _init_dir_and_version():
    import os
    path = os.path.abspath(os.path.dirname(__file__))
    try:
        import importlib.metadata
        version = importlib.metadata.version("geonum")
    except ModuleNotFoundError:
        from pkg_resources import get_distribution
        version = get_distribution('geonum').version
    return (path, version)



# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import numpy as np
from scipy.interpolate import interp1d

from geonum.exceptions import IntersectNotFound
from geonum.geopoint import GeoPoint
from geonum.lineongrid import LineOnGrid
from geonum.topodata import TopoData

class ElevationProfile(object):
    """Class for calculating elevation profiles

    This class can be used to compute elevation profiles for arbitrary
    locations on earth. This is done by using topographic data which is
    provided (instance of :class:`geonum.TopoData`). Profiles are computed with
    respect to input location (:attr:`observer`) and the azimuthal direction of
    the profile is specified via the provided endpoint location (
    :attr:`endpoint`).

    The profile is calculated between the two :class:`GeoPoint` objects using a
    provided topographic data grid which must cover the range spanned by
    both points.

    Note
    ----
    For class attributes see properties below.

    Parameters
    ----------
    observer : GeoPoint
        start point of profile
    endpoint : GeoPoint
        end point of profile
    topo_data : TopoData
        topographic dataset
    calc_on_init : bool, optional
        if True, then the profile is calculated on class inialisation.
        Default is True.
    **kwargs
        additional keyword args passed to :func:`det_profile` (only relevant
        if `calc_on_init` is True)

    Example
    -------
    >>> # define start and end location of profile
    >>> startpoint = GeoPoint(37.73122, 15.1129, name="Scientist")
    >>> endpoint = GeoPoint(37.751005, 14.993435, name="Etna Summit")
    >>> # Create profile retrieval engine
    >>> ep = ElevationProfile(observer=startpoint, endpoint=endpoint)
    >>> # Retrieve profile (uses default topographic data SRTM)
    >>> altitudes = ep.det_profile(interpolate=False)
    >>> # Print retrieved altitude levels along profile
    >>> print(altitudes)
    """
    def __init__(self, observer, endpoint, topo_data=None,
                 calc_on_init=True, **kwargs):

        self._topo_data = None
        self._observer = None
        self._endpoint = None

        self._check_set_observer(observer)
        self._check_set_endpoint(endpoint)

        if topo_data is not None:
            self.topo_data = topo_data

        self._init_attrs()

        if calc_on_init:
            self.det_profile(**kwargs)

    @property
    def observer(self) -> GeoPoint:
        """
        Start coordinate of elevation profile
        """
        return self._observer

    @observer.setter
    def observer(self, value):
        self._check_set_observer(value)
        self._init_attrs()

    @property
    def endpoint(self) -> GeoPoint:
        """
        End coordinate of elevation profile
        """
        return self._endpoint

    @endpoint.setter
    def endpoint(self, value):
        self._check_set_endpoint(value)
        self._init_attrs()

    @property
    def topo_data(self) -> TopoData:
        """
        Topographic data used to extract elevation profile

        Note
        ----
        Attempts automatic retrieval of topographic data if the data is not
        set.
        """
        if self._topo_data is None:
            self._topo_data = self._get_topo()
        return self._topo_data

    @topo_data.setter
    def topo_data(self, value):
        self._check_set_topo_data(value)
        self._init_attrs()

    @property
    def line(self) -> LineOnGrid:
        """
        Connection line between :attr:`observer` and :attr:`endpoint`

        Note
        ----
        private attribute that cannot be set by the user but is calculated
        automatically based on attributes :attr:`observer`, :attr:`endpoint`
        and :attr:`topo_data`.
        """
        if not isinstance(self._line, LineOnGrid):
            if self._coords_topo is None:
                self._set_latlon_indices_topogrid()
            self._line = LineOnGrid(**self._coords_topo)
        return self._line

    @property
    def observer_topogrid(self) -> GeoPoint:
        """
        Location of endpoint on topogrid (depends on topo resolution)

        Note
        ----
        Cannot be set by the user but is calculated
        automatically based on attributes :attr:`observer` and
        :attr:`topo_data`.
        """
        if self._observer_topogrid is None:
            if self._coords_topo is None:
                self._set_latlon_indices_topogrid()
            self._observer_topogrid = GeoPoint(
                self.topo_data.lats[self._coords_topo['y0']],
                self.topo_data.lons[self._coords_topo['x0']],
                topo_data=self.topo_data)
        return self._observer_topogrid

    @property
    def endpoint_topogrid(self) -> GeoPoint:
        """
        Location of endpoint on topogrid (depends on topo resolution)

        Note
        ----
        Cannot be set by the user but is calculated
        automatically based on attributes :attr:`endpoint` and
        :attr:`topo_data`.
        """
        if self._endpoint_topogrid is None:
            if self._coords_topo is None:
                self._set_latlon_indices_topogrid()
            self._endpoint_topogrid = GeoPoint(
                self.topo_data.lats[self._coords_topo['y1']],
                self.topo_data.lons[self._coords_topo['x1']],
                topo_data=self.topo_data)
        return self._endpoint_topogrid

    @property
    def dist_hor(self) -> float:
        """
        Horizontal distance in km between start and endpoint
        """
        return (self.endpoint_topogrid - self.observer_topogrid).dist_hor


    @property
    def azimuth(self) -> float:
        """
        Azimuth angle of profile (wrt to North direction)
        """
        return (self.endpoint_topogrid - self.observer_topogrid).azimuth

    @property
    def profile(self) -> np.ndarray:
        """
        Retrieved altitude levels in m along :attr:`line`

        Raises
        ------
        AttributeError
            if profile has not been calculated yet
        """
        if self._profile is None:
            raise AttributeError(
                'Profile information not available, call det_profile first')
        return self._profile

    @property
    def dists(self) -> np.ndarray:
        """
        Distances from observer in km along :attr:`line`

        Raises
        ------
        AttributeError
            if profile has not been calculated yet
        """
        if self._dists is None:
            raise AttributeError(
                'Distance information not available, call det_profile first')
        return self._dists

    @property
    def resolution(self) -> float:
        """Horizontal profile resolution (averaged from distance array)

        Note
        ----
        Only works if profile was already determined
        """
        return abs((self.dists[1:] - self.dists[:-1]).mean())

    @property
    def gradient(self) -> np.ndarray:
        """Gradient of profile at each point
        """
        return np.gradient(self.profile)

    @property
    def slope(self) -> np.ndarray:
        """Slope of profile at each point

        The slope is calculated for each profile point by dividing the
        associated altitude step by the corresponding horitontal step.
        """
        return np.gradient(self.profile)/(np.gradient(self.dists)*1000)

    @property
    def min(self) -> float:
        """Minimum altitude in elevation profile"""
        return np.nanmin(self.profile)

    @property
    def max(self) -> float:
        """Maximum altitude in elevation profile"""
        return np.nanmax(self.profile)

    @property
    def alt_range(self) -> float:
        """Altitude range of profile"""
        return self.max - self.min

    def _get_topo(self) -> TopoData:
        """
        Load topographic data between start and end point

        Returns
        -------
        TopoData
        """
        return self.observer.get_topo_data(geo_point=self.endpoint)[0]

    def slope_angles(self, decimal_degrees=True):
        """
        Get slope angles of profile (in each sample point)

        Parameters
        ----------
        decimal_degrees : bool
            if True, angles are converted from radians to degrees

        Returns
        -------
        np.ndarray
            slope angles at each profile point
        """
        a = np.tan(self.slope)
        if decimal_degrees:
            a = np.rad2deg(a)
        return a

    def slope_angle(self, dist):
        """
        Get slope angle of profile at input distance from observer

        Parameters
        ----------
        dist : float
            distance from observer in km for which slope is to be retrieved

        Returns
        -------
        float
            retrieved slope at input distance.
        """
        if not dist >= 0 or dist > np.nanmax(self.dists):
            raise ValueError('invalid input: dist must be positive')
        idx = np.argmin(abs(self.dists - dist))
        return self.slope_angles()[idx]

    def det_profile(self, interpolate=True, resolution=5.0, itp_type="linear",
                    **mapping_opts):
        """
        Determine the elevation profile

        Searches the closest tiles in the topo data grid for both observer and
        endpoint and based on these two points the elevation profile is
        extracted using a :class:`LineOnGrid` object (which extracts altitudes
        from the topo data along the connection vector of the 2 points using
        2D spline intepolation)

        Parameters
        ----------
        interpolate : bool
            if True, the profile is interpolated to a certain horizontal
            resolution
        resolution : float
            desired grid resolution in m for interpolation. Interpolation is
            performed if :attr:`interpolate` is True and if the actual
            resolution of the topo data is smaller than input, else, nothing
            is done
        itp_type : str
            interpolation type (e.g. "linear", "cubic")
        **mapping_opts
            additional keyword arguments that are passed to
            :func:`LineOnGrid.get_line_profile`

        Returns
        -------
        np.ndarray
            the array containing the retrieved altitude levels along the
            retrieval direction (from the observer)
        """
        line = self.line
        topo = self.topo_data
        altitudes = line.get_line_profile(topo.data, **mapping_opts)

        dists = np.linspace(0, self.dist_hor, line.length + 1)
        if interpolate:
            dists, altitudes = self._interp_helper(dists, altitudes,
                                                   resolution, itp_type)
        self._dists = dists
        self._profile = altitudes
        return altitudes

    def _interp_helper(self, dists, altitudes, resolution, itp_type):
        """
        Helper to interpolate profile

        Parameters
        ----------
        dists : np.ndarray
            Array with distances from observer
        altitudes : np.ndarray
            Array with retrieved altitudes at input distances
        resolution : int or float
            desired horizontal resolution in m
        itp_type : str
            interpolation type (passed to :func:`scipy.interpolate.interp1d`)

        Returns
        -------
        np.ndarray
            interpolated distances from observer.
        np.ndarray
            interpolated altitude levels.
        """
        res0 = (dists[1] - dists[0]) * 1000
        fac = int(np.ceil(res0 / resolution))
        if fac > 1:
            fz = interp1d(dists, altitudes, kind=itp_type)
            dists = np.linspace(0, self.dist_hor, self.line.length * fac)
            altitudes = fz(dists)
        else:
            print('No interpolation of elevation profile needed, data already '
                  'has sufficient resolution')
        return dists, altitudes

    def get_altitudes_view_dir(self, elev_angle, view_above_topo_m=0):
        """
        Get vector containing altitudes for a viewing direction

        The viewing direction is specified by the azimuth angle of the
        connection vector between observer and endpoint, the elevation angle
        needs to be specified via the input parameters, and the start point
        (first elevation value) corresponds to the altitude at the observer
        position plus an offset in m which can be specified.

        Parameters
        ----------
        elev_angle : float
            elevation angle of viewing direction
        view_above_topo_m : float
            altitude offset of start point in m, defaults to 0.

        Returns
        -------
        np.ndarray
            vector with altitude values (same length as ``self.profile``)

        """
        dh = 1000 * np.tan(np.radians(elev_angle))*self.dists
        result = dh+self.profile[0] + view_above_topo_m
        return result

    def get_first_intersection(self, elev_angle,view_above_topo_m=1.5,
                               min_dist=None,local_tolerance=3,
                               max_diff=None, plot=False):

        """Finds first intersection of a viewing direction with topography

        Start point of the viewing vector is the observer position (or more
        accurately, the center position of the closest topography tile in
        the topography data set). The relative altitude of the start point
        with respect to the topography altitude at the observer position
        can be controlled on input (arg ``view_above_topo_m``) as well as
        the elevation angle of the viewing direction. The azimuth angle
        corresponds to the profile azimuth (i.e. azimuth between
        :attr:`observer` and :attr:`endpoint` of this profile).

        The signal analysed for finding the intersection is a vector
        containing relative elevation positions of the viewing direction
        with respect to the profile altitude::

            diff_signal = self.get_altitudes_view_dir - self.profile

        The first intersection of the viewing direction with the topography
        is identified by the first zero crossing of this ``diff_signal``.

        Parameters
        ----------
        elev_angle : float
            elevation angle of viewing direction (in decimal degrees)
        view_above_topo_m : float
            altitude offset in m relative to altitude of start point (
            :attr:`observer`) in m. Defaults to 1.5 m.
        min_dist : float
            minimum distance (in km) of first considered intersection with
            topography from observer. If None, use 1% of distance between
            :attr:`observer` and :attr:`endpoint`.
        local_tolerance : int
            tolerance factor to estimate distance uncertainty based on topo
            grid resolution. Defaults to 3.
        max_diff : float, optional
            maximum allowed altitude difference in m between vector specifying
            viewing direction and the topography. Default to None, in which
            case the value assigned to :attr:`resolution` x 1000.
        plot : bool
            If true, the profile is plotted, including the intersection
            result.

        Returns
        -------
        float
            retrieved horizontal distance of intersection point.
        float
            error of retrieved horizontal distance of intersection point.
        GeoPoint
            location of intersect
        np.ndarray
            elevations along viewing direction vector arising from input
            elevation and altitude offset as well as the azimuth of the
            profile
        Axes or None
            matplotlib axes instance if input arg `plot` is True, else None.
        """

        if min_dist is None:
            min_dist = self.dist_hor*.01
        if max_diff is None:
            max_diff = self.resolution * 1000
        view_elevations = self.get_altitudes_view_dir(elev_angle,
                                                      view_above_topo_m)

        # determine the difference signal
        diff_signal = view_elevations - self.profile
        # First condition: consider only points in certain distance from observer
        cond1 = self.dists > min_dist
        # Second condition: consider only "close to zero" points (this might be
        # redundant, I'll leave it for now because it works)
        cond2 = abs(diff_signal) < max_diff

        # create array with all distances matching the 2 conditions
        dists_0 = self.dists[cond1 * cond2]
        # relax condition 2 if nothing was found
        if len(dists_0) == 0:
            raise IntersectNotFound(
                'could not establish initial array of candidate distances '
                'for retrieval of intersection point, you might succeed by '
                'setting or increasing the value of input parameter max_diff')

        try:
            # get all diff vals matching the 2 conditions
            diff_vals = diff_signal[cond1 * cond2]
            # get the index of the first zero crossing
            first_idx = np.where(np.diff(np.sign(diff_vals)))[0][0]

            # get distance and local tolerance value
            dist, tol = dists_0[first_idx], self.resolution * local_tolerance
            # make mask to access all distances within tolerance range
            cond4 = np.logical_and(dist - tol <= dists_0, dist + tol >= dists_0)
            # estimate distance error from standard deviation of all
            # distances within tolerance range
            dist_err = dists_0[cond4].std()

            # create geopoint corresponding to intersection
            intersect = self._observer_topogrid.offset(self.azimuth, dist)

            # set the altitude of this point (retrieved from profile)
            intersect.altitude = self.get_altitude_at_distance(dist) # / 1000.

        except IndexError:
            from traceback import format_exc
            raise IntersectNotFound(
                f'No intersections could be detected, traceback:\n'
                f'{format_exc()}')

        ax = None
        if plot:
            ax = self._plot_intersect_search_result(view_elevations, dist)
            ax.set_title("Azim: %.1f, Elev: %.1f, Intersect @ "
                "dist =%.1f km" %(self.azimuth, elev_angle, dist))

        return dist, dist_err, intersect, view_elevations, ax

    def find_horizon_elev(self, elev_start=0.0, elev_stop=60.0, step_deg=0.1,
                          **kwargs):
        """Find first elevation angle which does not intersect with topo

        Scans towards zenith elevation within input elevation range
        in elevation steps provided via input arg `step_deg`, starting at
        `elev_start`. For each elevation, an attempt is made to find an
        intersection of the resulting viewing direction vector with the local
        topography using :func:`get_first_intersection`. The horizon
        elevation is identified as the first elevation angle for which no
        intersection can be found.

        Parameters
        ----------
        elev_start : float
            start search elevation angle
        elev_stop : float
            stop search elevation angle
        step_deg : float
            angle step for search (coarser is faster)
        **kwargs:
            additional keyword agruments passed to
            :func:`get_first_intersection`

        Raises
        ------
        IntersectNotFound
            if no intersection with topography can be found for input
            elevation range.

        Returns
        -------
        float
            detected elevation angle of horizon
        list
            list of elevation angles for which intersections with topography
            could be identified.
        list
            corresponding horizontal distances of :attr:`observer` to the
            terrain features.
        """
        elevs = np.arange(elev_start, elev_stop + step_deg, step_deg)
        elev_sects = []
        dists_sects = []

        prev_elev = None
        for elev in elevs:
            try:
                (dist,
                 dist_err,
                 intersect,
                 view_elevations,_) = self.get_first_intersection(elev,
                                                                  plot=False,
                                                                  **kwargs)
            except IntersectNotFound:
                if prev_elev is None:
                    raise IntersectNotFound(
                        f'failed to determine elevation angle of horizon '
                        f'consider starting at a lower elevation than '
                        f'{elev_start}')
                else:
                    return (elev, elev_sects, dists_sects)
            dists_sects.append(dist), elev_sects.append(elev)
            prev_elev = elev

    def get_altitude_at_distance(self, dist):
        """Get altitude at a certain distance from observer

        Parameters
        ----------
        dist : float
            horizontal distance from :attr:`observer` along profile

        Returns
        -------
        float
            altitude at input distance.
        """
        idx = np.argmin(abs(self.dists - dist))
        return self.profile[idx]

    def plot(self, ax=None, facecolor=None, alpha=0.20):
        """Plot the elevation profile
        Parameters
        ----------
        ax : Axes, optional
            matplotlib axes object to be used for plot
        facecolor : str, optional
            color of profile, defaults to "#994d00" (light beige)
        alpha : float, optional
            opaqueness of profile, defaults to 0.20.

        Returns
        -------
        Axes
            matplotlib axes instance used for plot
        """
        if facecolor is None:
            facecolor = "#994d00"
        if ax is None:
            from matplotlib.pyplot import subplots
            fig, ax = subplots(1,1)
        dists = self.dists
        ax.fill_between(dists, self.profile, facecolor=facecolor,
                        alpha=alpha)
        ax.set_xlabel("Distance [km]")
        ax.set_ylabel("Altitude [m]")
        ax.set_ylim([self.min - .1 * self.alt_range,
                     self.max + .1 * self.alt_range])
        ax.set_xlim([0, dists[-1]])
        return ax

    def __call__(self, dist):
        """Returns altitude at a certain distance

        Wrapper for :func:`get_altitude_at_distance`

        Parameters
        ----------
        dist : float
            distance in km

        Returns
        -------
        float
            altitude at input distance.
        """
        return self.get_altitude_at_distance(dist)

    def _init_attrs(self):
        """
        Initiate private attributes required in this class
        """
        self._observer_topogrid = None
        self._endpoint_topogrid = None
        self._coords_topo = None
        self._line = None
        self._profile=None
        self._dists=None

    def _check_set_observer(self, value):
        """Setter helper for :attr:`observer`

        Parameters
        ----------
        value : GeoPoint
            observer candidate

        Raises
        ------
        ValueError
            if input is not instance of :class:`GeoPoint`

        """
        if not isinstance(value, GeoPoint):
            raise ValueError('Need instance of geonum.GeoPoint')
        self._observer = value

    def _check_set_endpoint(self, value):
        """Setter helper for :attr:`endpoint`

        Parameters
        ----------
        value : GeoPoint
            endpoint candidate

        Raises
        ------
        ValueError
            if input is not instance of :class:`GeoPoint`
        """
        if not isinstance(value, GeoPoint):
            raise ValueError('Need instance of geonum.GeoPoint')
        self._endpoint = value

    def _check_set_topo_data(self, value):
        """Setter helper for :attr:`topo_data`

        Parameters
        ----------
        value : TopoData
            topo_data candidate

        Raises
        ------
        ValueError
            if input is not instance of :class:`TopoData`
        """
        if not isinstance(value, TopoData):
            raise ValueError('Need instance of geonum.TopoData')
        self._topo_data = value

    def _set_latlon_indices_topogrid(self):
        """Calculate and set indeices of start and stop point on topogrid"""
        topo = self.topo_data
        self._coords_topo = {
         'x0' : np.argmin(abs(topo.lons - self.observer.lon.decimal_degree)),
         'y0' : np.argmin(abs(topo.lats - self.observer.lat.decimal_degree)),
         'x1' : np.argmin(abs(topo.lons - self.endpoint.lon.decimal_degree)),
         'y1' : np.argmin(abs(topo.lats - self.endpoint.lat.decimal_degree))
        }

    def _plot_intersect_search_result(self, view_elevations, dist=None):
        """
        Plot result from intersection point search

        Parameters
        ----------
        view_elevations : list-like
            elevation levels along line of sight

        dist : float, optional
            distance to intersect

        Returns
        -------
        Axes
            matplotlib axes instance used for plotting

        """
        ax = self.plot()
        ax.plot(self.dists, view_elevations, label = "Viewing direction")
        if dist is not None:
            ax.axvline(dist, ls = "--", label = "Intersection")

        ax.legend(loc="best", fancybox=True, framealpha=0.4)
        return ax



# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import abc
import os
from warnings import warn

import numpy as np
import srtm

from geonum import NETCDF_AVAILABLE, LOCAL_TOPO_DIR
from geonum.exceptions import (TopoAccessError, SRTMNotCoveredError)


class TopoAccessBase(abc.ABC):
    """Abstract base class for topgraphy file implementations

    Defines minimum interface for derived access classes of different
    topographic datasets.
    
    """
    #: A coordinate for which data should be available
    _TESTLAT = 45
    _TESTLON = 15

    @abc.abstractmethod
    def get_data(self, lat0, lon0, lat1=None, lon1=None):
        """Declaration of data access method

        It is obligatory to implement this method into derived classes.

        Parameters
        ----------
        lat0 : float
            first latitude coordinate of topographic range (lower left coord)
        lon0 : float
            first longitude coordinate of topographic range (lower left coord)
        lat1 : int or float, optional
            second latitude coordinate of topographic range (upper right
            coord). If None only data around lon0, lat0 will be extracted.
        lon1 : int or float, optional
            second longitude coordinate of topographic range (upper right
            coord). If None only data around lon0, lat0 will be extracted.

        Returns
        -------
        TopoData
            instance of TopoData class
        """
        pass # pragma: no cover

    def check_access(self):
        """Check if topography data can be accessed"""
        try:
            self.get_data(self._TESTLAT, self._TESTLON)
            return True
        except Exception as e:
            print('Could not access topodata: {}'.format(repr(e)))
        return False

    def _prep_borders(self, lat0, lon0, lat1, lon1):
        """Sort by longitudes and determines LL and TR coordinates

        Parameters
        ----------
        lat0 : float
            first latitude coordinate of topographic range (lower left coord)
        lon0 : float
            first longitude coordinate of topographic range (lower left coord)
        lat1 : float
            second latitude coordinate of topographic range (upper right coord)
        lon1 : float
            second longitude coordinate of topographic range (upper right coord)

        Returns
        -------
        tuple
            4-element tuple, containing:

            - float, smallest latitude (LL corner)
            - float, smallest longitude (LL corner)
            - float, largest latitude (TR corner)
            - float, largest longitude (TR corner)
        """
        lats, lons = np.asarray([lat0, lat1]), np.asarray([lon0, lon1])
        return (np.nanmin(lats), np.nanmin(lons),
                np.nanmax(lats), np.nanmax(lons))

    def _init_lons_lats(self, lats_all, lons_all, lat0, lon0, lat1=None,
                        lon1=None):
        """Get all latitudes and longitudes on a topodata grid

        Parameters
        ----------
        lats_all : ndarray
            numpy array containing available latitudes of the accessed topo
            dataset
        lons_all : ndarray
            numpy array containing available longitudes of the accessed topo
            dataset
        lat0 : float
            first latitude coordinate of topographic range (lower left coord)
        lon0 : float
            first longitude coordinate of topographic range (lower left coord)
        lat1 : float, optional
            second latitude coordinate of topographic range (upper right coord)
        lon1 : float, optional
            second longitude coordinate of topographic range (upper right coord)

        Returns
        -------
        tuple
            2-element tuple, containing

            - ndarray, topodata latitudes overlapping with input range
            - ndarray, topodata longitudes overlapping with input range
        """
        if any([x is None for x in [lat1, lon1]]):
            lat1, lon1 = lat0, lon0

        if lon0 > lon1:
            lon0, lon1 = lon1, lon0
        if lat0 > lat1:
            lat0, lat1 = lat1, lat0

        # closest indices
        idx_lons = [np.argmin(abs(lons_all - lon0)),
                    np.argmin(abs(lons_all - lon1))]
        idx_lats = [np.argmin(abs(lats_all - lat0)),
                    np.argmin(abs(lats_all - lat1))]

        # Make sure that the retrieved indices actually INCLUDE the input ranges
        if idx_lons[0] == 0 and lons_all[0] > lon0:
            warn("Error: Lon0 smaller than range covered by file, using first"
                 " available index in topodata..")
            idx_lons[0] = 0
        elif lons_all[idx_lons[0]] > lon0:
            idx_lons[0] -= 1
        if idx_lons[1] == len(lons_all) - 1 and lons_all[-1] < lon1:
            warn("Error: Lon1 larger than range covered by file, using last"
                 " available index in topodata..")
            idx_lons[1] = len(lons_all) - 1
        elif lons_all[idx_lons[1]] < lon1:
            idx_lons[1] += 1
        if idx_lats[0] == 0 and lats_all[0] > lat0:
            warn("Error: Lat0 smaller than range covered by file, using first"
                 " available index in topodata..")
            idx_lats[0] = 0
        elif lats_all[idx_lats[0]] > lat0:
            idx_lats[0] -= 1
        if idx_lats[1] == len(lats_all) - 1 and lats_all[-1] < lat1:
            warn("Error: Lat1 larger than range covered by file, using last"
                 " available index in topodata..")
            idx_lats[1] = len(lats_all) - 1
        elif lats_all[idx_lats[1]] < lat1:
            idx_lats[1] += 1

        return (lats_all[idx_lats[0]: idx_lats[1] + 1],
                lons_all[idx_lons[0]: idx_lons[1] + 1],
                idx_lats, idx_lons)


class Etopo1Access(TopoAccessBase):
    """A class representing netCDF4 data access of Etopo1 data

    See `here <https://github.com/jgliss/geonum#supported-etopo1-files>`_ for
    instructions on the data access.

    Attributes
    ----------
    loader
        data loader (:class:`netCDF4.Dataset`)

    Parameters
    ----------
    local_path : str
        directory where Etopo1 data files are stored
    file_name :  str
        file name of etopo data file
    check_access : bool
        if True, then access to topography data is checked on init and an
        error is raised if no dataset can be accessed
    search_database : bool
        if True and topodata file :attr:`file_path` does not exist, then
        a valid topography file is searched in all paths that are specified
        in file `~/.geonum/LOCAL_TOPO_PATHS`.

    Raises
    ------
    TopoAccessError
        if input arg `check_access` is True and if no supported data file
        can be found

    """
    #: ID of dataset
    topo_id = "etopo1"

    #: filenames of supported topographic datasets in preferred order
    supported_topo_files = ["ETOPO1_Ice_g_gmt4.grd",
                            "ETOPO1_Bed_g_gmt4.grd"]

    def __init__(self, local_path=None, file_name=None, check_access=False,
                 search_database=True):
        if file_name is None:
            file_name = "ETOPO1_Ice_g_gmt4.grd"
        if not NETCDF_AVAILABLE: # pragma: no cover
            raise ModuleNotFoundError(
                "Etopo1Access class cannot be initiated. "
                "Please install netCDF4 library first")
        if local_path is None:
            local_path = LOCAL_TOPO_DIR
        self._local_path = local_path
        self._file_name = file_name

        from netCDF4 import Dataset

        self.loader = Dataset

        if not os.path.exists(self.file_path) and search_database:
            self.search_topo_file_database()

        # check if file exists
        if check_access:
            if not os.path.exists(self.file_path):
                raise TopoAccessError('File {} could not be found in local '
                                      'topo directory: {}'.format(
                    self.file_name,
                    self.local_path))
            elif not self.check_access():
                raise TopoAccessError('Failed to extract topography data for '
                                      'Etopo dataset')

    @property
    def local_path(self):
        """Directory containing ETOPO1 gridded data files"""
        return self._local_path

    @local_path.setter
    def local_path(self, val):
        if not os.path.exists(val) or not os.path.isdir(val):
            raise ValueError(f'Input directory {val} does not exist or is not '
                             f'a directory...')
        self._check_topo_path(val)
        self._local_path = val

    @property
    def file_name(self):
        """File name of topographic dataset used"""
        return self._file_name

    @file_name.setter
    def file_name(self, val):
        if not val in self.supported_topo_files:
            raise ValueError(
                f'Invalid file name for Etopo1 dataset {val}. Valid filenames '
                f'are: {self.supported_topo_files}')
        self._file_name = val

    @property
    def file_path(self):
        """Return full file path of current topography file"""
        return os.path.join(self.local_path, self.file_name)

    @file_path.setter
    def file_path(self, val):
        self.set_file_location(val)

    def _check_topo_path(self, path):
        """Check if path exists and if it is already included in database

        Parameters
        ----------
        path : str
            path to be checked
        """
        from geonum.helpers import check_and_add_topodir
        check_and_add_topodir(path)

    def _get_all_local_topo_paths(self):
        """Get all search paths for topography files"""
        from geonum.helpers import all_topodata_search_dirs
        return all_topodata_search_dirs()

    def _search_topo_file(self, path=None):
        """Checks if a valid etopo data file can be found in local folder

        Searches in ``self.local_path`` for any of the file names specified
        in ``supported_topo_files``

        """
        if path is None:
            path = self.local_path
        print(f'Searching valid topo file in folder: {path}')
        fnames = os.listdir(path)
        for name in fnames:
            if name in self.supported_topo_files:
                self.file_name = name
                self.local_path = path
                print(("Found match, setting current filepath: %s"
                       % self.file_path))
                return True
        return False

    def _find_supported_files(self):
        """Look for all supported files in ``self.local_path```and return list"""
        files = os.listdir(self.local_path)
        lst = []
        for name in files:
            if name in self.supported_topo_files:
                lst.append(name)
        return lst

    def search_topo_file_database(self):
        """Checks if a valid topo file can be found in database"""
        all_paths = self._get_all_local_topo_paths()
        for path in all_paths:
            if self._search_topo_file(path):
                return True
        return False

    def get_data(self, lat0, lon0, lat1=None, lon1=None):
        """Retrieve data from topography file

        Parameters
        ----------
        lat0 : float
            first latitude coordinate of topographic range (lower left coord)
        lon0 : float
            first longitude coordinate of topographic range (lower left coord)
        lat1 : int or float, optional
            second latitude coordinate of topographic range (upper right
            coord). If None only data around lon0, lat0 will be extracted.
        lon1 : int or float, optional
            second longitude coordinate of topographic range (upper right
            coord). If None only data around lon0, lat0 will be extracted.

        Returns
        -------
        TopoData
            instance of TopoData class
        """
        from geonum import TopoData
        etopo1 = self.loader(self.file_path)

        lons = etopo1.variables["x"][:]
        lats = etopo1.variables["y"][:]

        lats, lons, idx_lats, idx_lons = self._init_lons_lats(lats, lons, lat0,
                                                              lon0, lat1, lon1)

        vals = np.asarray(etopo1.variables["z"][idx_lats[0]: idx_lats[1] + 1,
                          idx_lons[0]: idx_lons[1] + 1],
                          dtype=float)
        etopo1.close()

        return TopoData(lats, lons, vals, data_id=self.topo_id)


class SRTMAccess(TopoAccessBase):
    """Class for SRTM topographic data access

    Uses library `srtm.py <https://pypi.python.org/pypi/SRTM.py/0.3.1>`_
    for online access of data.

    Note
    ----
    :mod:`srtm.py` downloads the topo data from `this source <http://
    dds.cr.usgs.gov/srtm/version2_1/>`_ and stores a copy of the unzipped data
    files in the current cache directory found in home.

    Whenever data access is requested, the :mod:`srtm.py` checks if the file
    already exists on the local machine and if not downloads it online. The
    online access is rather slow, so do not be surprised, if things take a
    while when accessing a specific location for the first time.

    **Deleting cached SRTM files**:
        use :func:`geonum.topoaccess.delete_all_local_srtm_files`

    Parameters
    ----------
    check_access : bool
        check if data can be accessed on class initialisation
    **kwargs
        additional keyword arguments that are passed through (irrelevant for
        this class but relevant for factory loader class
        :class:`TopoDataAccess`, particularly :func:`set_mode` therein.
    """

    def __init__(self, check_access=False, **kwargs):
        """Class initialisation"""
        self.loader = srtm
        self.topo_id = "srtm"

        if check_access:
            self.check_access()

    def _coordinate_covered(self, access_obj, lat, lon):
        """Checks if SRTM data is available for input coordinate

        Parameters
        ----------
        access_obj : GeoElevationData
            data access object from :mod:`srtm.py` module
            (can be created calling ``srtm.get_data()``)
        lat : float
            latitude of point
        lon : float
            longitude of point

        Returns
        -------
        bool
            True, if SRTM data is available for coordinate, else False.
        """
        if access_obj.get_file_name(lat, lon) is None:
            return False
        return True

    def get_data(self, lat0, lon0, lat1=None, lon1=None):
        """Load SRTM topographic subset for input range

        Parameters
        ----------
        lat0 : float
            first latitude coordinate of topographic range (lower left coord)
        lon0 : float
            first longitude coordinate of topographic range (lower left coord)
        lat1 : int or float, optional
            second latitude coordinate of topographic range (upper right
            coord). If None only data around lon0, lat0 will be extracted.
        lon1 : int or float, optional
            second longitude coordinate of topographic range (upper right
            coord). If None only data around lon0, lat0 will be extracted.

        Returns
        -------
        TopoData
            instance of TopoData class
        """
        from geonum import TopoData
        print("Retrieving SRTM data (this might take a while) ... ")
        # create GeoElevationData object for data access
        dat = self.loader.get_data()
        # check if second input point is specified and set equal first point if
        # not
        if any([x is None for x in [lat1, lon1]]):
            lat1, lon1 = lat0, lon0
        # Check if first point is covered by dataset
        if not self._coordinate_covered(dat, lat0, lon0):
            raise SRTMNotCoveredError('Point (lat={:.2f}, lon={:.2f}) not '
                                      'covered by SRTM'.format(lat0, lon0))
        # check if second point is covered by dataset
        if not self._coordinate_covered(dat, lat1, lon1):
            raise SRTMNotCoveredError('Endpoint coordinate (lat={:.2f}, '
                                      'lon={:.2f}) not covered by SRTM'
                                      .format(lat1, lon1))
        # prepare borders of covered lon / lat regime
        lat_ll, lon_ll, lat_tr, lon_tr = self._prep_borders(lat0, lon0,
                                                            lat1, lon1)
        # get SRTM file for lower left corner of regime
        f_ll = dat.get_file(lat_ll, lon_ll)
        # get SRTM file for top right corner of regime
        f_tr = dat.get_file(lat_tr, lon_tr)
        # create array of longitude values for regime
        lons_all = np.linspace(f_ll.longitude, f_tr.longitude + 1,
                               f_ll.square_side)
        # create array of latitude values for regime
        lats_all = np.linspace(f_ll.latitude, f_tr.latitude + 1,
                               f_ll.square_side)
        # prepare coordinates
        lats, lons, _, _ = self._init_lons_lats(lats_all, lons_all,
                                                lat0, lon0, lat1, lon1)
        # Init data array
        vals = np.ones((len(lats), len(lons))) * np.nan
        # loop over all coordinates and try access the elevation data
        for i in range(len(lats)):
            for j in range(len(lons)):
                vals[i, j] = dat.get_elevation(lats[i], lons[j])

        return TopoData(lats, lons, vals, data_id=self.topo_id)



import numpy.testing as npt
import pytest
from geonum.conftest import skip_srtm, does_not_raise_exception
import numpy as np
from geonum.exceptions import IntersectNotFound
from geonum import GeoPoint, ElevationProfile, TopoData
from matplotlib.axes import Axes

LAT0, LON0 = -39.296571, 173.9224
LAT1, LON1 = -39.3538, 174.4383

P0 = GeoPoint(LAT0, LON0)
P1 = GeoPoint(LAT1, LON1)

@pytest.fixture(scope='module')
def profile():
    return ElevationProfile(P0, P1,
                            calc_on_init=False)

@pytest.fixture(scope='module')
def profile1():
    prof = ElevationProfile(P0, P1,
                            calc_on_init=False)
    prof.det_profile(**{'interpolate': True, 'resolution': 1000})
    return prof

@pytest.mark.filterwarnings("ignore:Failed to compute elevation profile.")
def test_ElevationProfile_wrong_input():
    with pytest.raises(ValueError):
        ElevationProfile('bla', 1, 'blub')

@pytest.mark.parametrize('args,raises', [
    ({'observer': 1}, pytest.raises(TypeError)),
    ({'observer': 1, 'endpoint': 2}, pytest.raises(ValueError)),
    ({'observer': GeoPoint(46,7),
      'endpoint': GeoPoint(46.1,7.1),
      'calc_on_init': False}, does_not_raise_exception())
    ])
def test_ElevationProfile___init__(args,raises):
    with raises:
        prof = ElevationProfile(**args)
        assert isinstance(prof, ElevationProfile)

def test_ElevationProfile_observer(profile):
    # test getter
    val = profile.observer
    assert isinstance(val, GeoPoint)
    # test setter
    profile.observer = val
    assert profile.observer is val

def test_ElevationProfile_endpoint(profile):
    # test getter
    val = profile.endpoint
    assert isinstance(val, GeoPoint)
    # test setter
    profile.endpoint = val
    assert profile.endpoint is val

def test_ElevationProfile_topo_data(profile):
    # test getter
    val = profile.topo_data
    assert isinstance(val, TopoData)
    # test setter
    profile.topo_data = val
    assert profile.topo_data is val

def test_ElevationProfile_observer_topogrid(profile):
    profile._coords_topo = None
    val = profile.observer_topogrid
    assert isinstance(val, GeoPoint)

def test_ElevationProfile_endpoint_topogrid(profile):
    profile._coords_topo = None
    val = profile.endpoint_topogrid
    assert isinstance(val, GeoPoint)

def test_ElevationProfile_azimuth(profile):
    val = profile.azimuth
    npt.assert_allclose(val, 98.3, atol=0.1)

def test_ElevationProfile_profile_unavail(profile):
    with pytest.raises(AttributeError):
        profile.profile

def test_ElevationProfile_dists_unavail(profile):
    with pytest.raises(AttributeError):
        profile.dists

@pytest.mark.parametrize('args,num,avg', [
    ({}, 9222, 576),
    ({'interpolate' : False},319, 575),
    ({'interpolate' : True, 'resolution' : 1000},319, 575),
    ])
def test_ElevationProfile_det_profile(profile, args, num, avg):
    val = profile.det_profile(**args)
    assert isinstance(val, np.ndarray)
    assert len(val) == num
    mean = np.nanmean(val)
    npt.assert_allclose(mean, avg, atol=1)

def test_ElevationProfile_dists_fail():
    with pytest.raises(AttributeError):
        p = ElevationProfile(P0,P1,calc_on_init=False)
        p.dists
    p.det_profile()
    assert isinstance(p.dists, np.ndarray)

def test_ElevationProfile_resolution(profile):
    profile.det_profile()
    res = profile.resolution
    assert isinstance(res, float)
    npt.assert_allclose(res, 0.005, atol=0.01)

def test_ElevationProfile_gradient(profile):
    profile.det_profile()
    grad = profile.gradient
    assert isinstance(grad, np.ndarray)
    assert len(grad) == 9222
    npt.assert_allclose(np.mean(grad), -0.015, atol=0.01)

def test_ElevationProfile_slope(profile):
    profile.det_profile()
    sl = profile.slope
    assert isinstance(sl, np.ndarray)
    npt.assert_allclose(np.mean(sl), -0.003, atol=0.001)

def test_ElevationProfile_min(profile1):
    npt.assert_allclose(profile1.min, 166.831, atol=0.01)

def test_ElevationProfile_max(profile1):
    npt.assert_allclose(profile1.max, 1858.777, atol=0.01)

def test_ElevationProfile_alt_range(profile1):
    npt.assert_allclose(profile1.alt_range, 1858.777-166.831, atol=0.01)

@pytest.mark.parametrize('decimal_degree,avg', [
    (False, -0.0027),
    (True, -0.158)
])
def test_ElevationProfile_slope_angles(profile1,decimal_degree,avg):
    sla = profile1.slope_angles(decimal_degree)
    npt.assert_allclose(np.nanmean(sla),avg,rtol=1e-1)

@pytest.mark.parametrize('dist,val,raises', [
    (1, 2.29, does_not_raise_exception()),
    (-1, 2.29, pytest.raises(ValueError)),
    (1999, 2.29, pytest.raises(ValueError))
])
def test_ElevationProfile_slope_angle(profile1,dist,val,raises):
    with raises:
        sla = profile1.slope_angle(dist)
        npt.assert_allclose(np.nanmean(sla),val,rtol=1e-1)

@pytest.mark.parametrize('elev_angle,view_above_topo_m,num,avg', [
    (0, 0, 319, 318),
    (0, 10, 319, 328),
    (2, 10, 319, 1114),
    ])
def test_ElevationProfile_get_altitudes_view_dir(profile1, elev_angle,
                                                 view_above_topo_m,num,avg):

    alts = profile1.get_altitudes_view_dir(elev_angle, view_above_topo_m)
    assert isinstance(alts, np.ndarray)
    assert len(alts) == num
    mean = np.mean(alts)
    npt.assert_allclose(mean, avg, atol=1)

@pytest.mark.parametrize(
    'elev_angle,view_above_topo_m,min_dist,local_tolerance,max_diff,plot,d,'
    'derr,alt,raises', [
    (0,0,0,3,0,True,31.1,0.3,319.2,pytest.raises(IntersectNotFound)),
    (0,0,0,3,None,True,31.1,0.3,319.2,does_not_raise_exception()),
    (0,0,0,10,None,True,31.1,0.9,319.2,does_not_raise_exception()),
    (6,0,0,10,None,True,10.8,0.7,1399.7,does_not_raise_exception()),
    (10,0,0,10,None,True,np.nan,np.nan,None,pytest.raises(IntersectNotFound))
    ]
    )
def test_ElevationProfile_get_first_intersection(profile,
    elev_angle,view_above_topo_m,min_dist,local_tolerance,max_diff,plot,
    d,derr,alt,raises):
    profile.det_profile(**{'interpolate': True, 'resolution': 1000})
    with raises:
        val = profile.get_first_intersection(elev_angle,view_above_topo_m,min_dist,
                                             local_tolerance,max_diff,plot)
        dist, dist_err, intersect, view_elevations, ax = val

        assert isinstance(view_elevations, np.ndarray)
        if np.isnan(d):
            assert intersect is None
            assert np.isnan(dist)
            assert np.isnan(dist_err)
        else:
            assert isinstance(intersect, GeoPoint)
            npt.assert_allclose([dist,dist_err,intersect.altitude], [d, derr, alt],
                            atol=0.1)
        if plot:
            assert isinstance(ax, Axes)

@pytest.mark.parametrize(
    'elev_start,elev_stop,step_deg,raises,num,elev', [
    (70,71,0.01,pytest.raises(IntersectNotFound),10,6.9),
    (0,10,0.1,does_not_raise_exception(),69,6.9),
    (6.8,7,0.01,does_not_raise_exception(),10,6.9),

    ])
def test_ElevationProfile_find_horizon_elev(profile1,
    elev_start,elev_stop,step_deg,raises,num,elev):
    with raises:
        (_elev,
         elev_sects,
         dists_sects) = profile1.find_horizon_elev(elev_start, elev_stop,
                                                  step_deg)

        npt.assert_allclose(_elev, elev, atol=0.1)
        assert len(elev_sects) == len(dists_sects) == num

@skip_srtm
def test_ElevationProfile_via_GeoPoint():


    obs = GeoPoint(LAT0, LON0,
                   auto_topo_access=True)

    npt.assert_almost_equal(obs.altitude, 324)

    prof = obs.get_elevation_profile(azimuth=90, dist_hor=50)

    prof_nans = obs.get_elevation_profile(azimuth=45, dist_hor=50)
    prof2 = obs.get_elevation_profile(azimuth=45, dist_hor=50, order=1)

    npt.assert_array_equal([len(prof.dists), len(prof.profile),
                            np.sum(np.isnan(prof_nans.profile))],
                            [10092,10092,10263])
    npt.assert_array_almost_equal([prof.profile.max(), prof.profile.min(),
                                   prof.profile.mean(), prof.profile.std(),
                                   prof.dists.max(),
                                   prof2.profile.max(), prof2.profile.min()],
                                  [2482.598405, 176.008484, 583.012202,
                                   486.57815, 50.033677, 1005.208932, 0.])

def test__check_set_endpoint_FAIL():
    ep = ElevationProfile(GeoPoint(45,15), GeoPoint(45,16))
    with pytest.raises(ValueError):
        ep._check_set_endpoint('42')

def test__check_set_topo_data_FAIL():
    ep = ElevationProfile(GeoPoint(45,15), GeoPoint(45,16))
    with pytest.raises(ValueError):
        ep._check_set_topo_data('42')

def test_ElevationProfile___call__(profile1):
    val = profile1(100)
    assert int(val) == 175




from geonum import GeoPoint, delete_local_srtm_files
from geonum.utils import get_local_srtm_files


def test_get_local_srtm_files():
    # make sure there is one file at least
    p = GeoPoint(45, 15, auto_topo_access=True)
    assert p.altitude > 0
    files = get_local_srtm_files()
    assert isinstance(files, list)
    assert len(files) >= 1


def test_delete_local_srtm_files():
    delete_local_srtm_files()
    assert len(get_local_srtm_files()) == 0


# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from os.path import exists
from warnings import warn

from matplotlib.pyplot import get_cmap
from numpy import asarray, nanmin, nanmax

from geonum import BASEMAP_AVAILABLE
from geonum.exceptions import OutOfDomain
from geonum.geopoint import GeoPoint
from geonum.geovector3d import GeoVector3D
from geonum.topodata import TopoData
from geonum.topodataaccess import TopoDataAccess

if BASEMAP_AVAILABLE:  # pragma: no cover
    from geonum.mapping import Map


class GeoSetup(object):
    """The GeoSetup class represents a collection of GeoPoints and vectors

    Attributes
    ----------
    id : str
        name of this setup
    points : dict
       contains :class:`GeoPoint` objects assigned to this setup
    vectors : dict
        contains :class:`GeoVector3D` objects assigned to this setup

    Parameters
    ----------
    points : list
        list of :class:`GeoPoint` objects to be included in this setup
    vectors : list
        list of :class:`GeoVector3D` objects to be included in this setup
    lat_ll : float, optional
        lower left latitude of regime
    lon_ll : float, optional
        lower left longitude of regime
    lat_tr : float, optional
        top right latitude of regime
    lon_tr : float, optional
        top right longitude of regime
    id : str
        identification string of this setup
    topo_access_mode : str
        topo data mode, default is SRTM
        (see :class:`TopoDataAccess` for details)
    local_topo_path : str
        local path were topography data (e.g. ETOPO1 data) is stored
    cmap_vecs : str
        String specifying a valid matplotlib colormap supposed to be
        used for drawing :class:`GeoVector3D` objects into overview
        maps
    init_borders : bool
        Whether or not lower left (ll) and top right (tr) border points are
        supposed to be created.

    """

    def __init__(self, points=None, vectors=None, lat_ll=None, lon_ll=None,
                 lat_tr=None, lon_tr=None, id=None,
                 topo_access_mode=None, local_topo_path=None,
                 cmap_vecs=None, init_borders=True):
        if id is None:
            id = 'MyGeoSetup'
        if topo_access_mode is None:
            topo_access_mode = "srtm"
        if cmap_vecs is None:
            cmap_vecs = 'Greens'

        self._cmap_vecs = cmap_vecs
        self.id = id
        self.points = {}
        self.vectors = {}

        self.topo_access_mode = topo_access_mode
        self.local_topo_path = local_topo_path

        self.topo_data = None
        if points is None:
            points = []
        elif isinstance(points, GeoPoint):
            points = [points]
        if not isinstance(points, list):
            raise ValueError('invalid input for points. need GeoPoint or list')

        for pt in points:
            if isinstance(pt, GeoPoint):
                self.add_geo_point(pt, assert_in_domain=False)

        if vectors is None:
            vectors = []
        elif isinstance(vectors, GeoVector3D):
            vectors = [vectors]
        if not isinstance(vectors, list):
            raise ValueError(
                'invalid input for points. need GeoVector3D or list')

        for vec in vectors:
            if isinstance(vec, GeoVector3D):
                self.add_geo_vector(vec)

        if lat_ll is not None:
            try:
                self.new_geo_point(lat_ll, lon_ll, name="ll")
                self.new_geo_point(lat_tr, lon_tr, name="tr")
            except (TypeError, ValueError):
                pass
        if init_borders and not self.borders_set and self.has_points():
            self.set_borders_from_points()

    @property
    def ll(self) -> GeoPoint:
        """Lower left coordinate of domain"""
        if not 'll' in self.points:
            raise AttributeError(
                'Lower left corner of GeoSetup domain is not defined'
            )
        return self.points["ll"]

    @ll.setter
    def ll(self, value):
        if not isinstance(value, GeoPoint):
            raise ValueError("Could not set lower left coordinate in "
                             "GeoSetup: need GeoPoint object")
        self.points["ll"] = value
        if self.has_topo_data():
            print('updated lower-left coordinate, consider reloading '
                  'topographic dataset using method load_topo_data for '
                  'updated domain.')

    @property
    def tr(self) -> GeoPoint:
        """Top right coordinate of domain"""
        if not 'tr' in self.points:
            raise AttributeError(
                'Top right corner of GeoSetup domain is not defined'
            )
        return self.points["tr"]

    @tr.setter
    def tr(self, value):
        if not isinstance(value, GeoPoint):
            raise ValueError("Could not set top right coordinate in "
                             "GeoSetup: need GeoPoint object")
        self.points["tr"] = value
        if self.has_topo_data():
            print('updated top-right coordinate, consider reloading '
                  'topographic dataset using method load_topo_data for '
                  'updated domain.')

    @property
    def lon_ll(self) -> float:
        """Longitude in decimal degrees of lower left coordinate of domain"""
        return self.ll.longitude

    @property
    def lat_ll(self) -> float:
        """Latitude in decimal degrees of lower left coordinate of domain"""
        return self.ll.latitude

    @property
    def lon_tr(self) -> float:
        """Longitude in decimal degrees of top right coordinate of domain"""
        return self.tr.longitude

    @property
    def lat_tr(self) -> float:
        """Latitude in decimal degrees of top right coordinate of domain"""
        return self.tr.latitude

    @property
    def delta_lon(self) -> float:
        """Longitude range of domain (in decimal degrees)"""
        return abs(self.lon_tr - self.lon_ll)

    @property
    def delta_lat(self) -> float:
        """Latitude range of domain (in decimal degrees)"""
        return abs(self.lat_tr - self.lat_ll)

    @property
    def borders_set(self) -> bool:
        """
        Boolean specifying whether domain borders are set or not
        """
        return True if self.has_point('ll') and self.has_point('tr') else False

    @property
    def center_coordinates(self) -> tuple:
        """Lat / Lon coordinates of center of domain"""
        return (self.lat_ll + self.delta_lat / 2.,
                self.lon_ll + self.delta_lon / 2.)

    @property
    def magnitude(self):
        """Returns dimension (in km) of area covered by this setup"""
        return (self.tr - self.ll).norm

    @property
    def topo_access(self):
        """Topography data access class"""
        return TopoDataAccess(self.topo_access_mode,
                              self.local_topo_path)

    @property
    def cmap_vecs(self):
        """Default colormap used for drawing vectors"""
        return get_cmap(self._cmap_vecs)

    def has_points(self):
        """
        Determine whether any points are set in this GeoSetup

        Returns
        -------
        bool
            True if one or more GeoPoints are registered, else False
        """
        if len(self.points) > 0:
            return True
        return False

    def has_topo_data(self):
        """
        Check whether topographic data is assigned or not

        Returns
        -------
        bool
            True, if topo data is available, else not

        """
        return True if isinstance(self.topo_data, TopoData) else False

    def set_local_topo_path(self, p):
        """Sets local path for Etopo1 data files can be found

        Note
        ----
        The default topo mode is "srtm" which provides online access, so
        it is not mandatory to provide topography data locally. However,
        the SRTM model has no global coverage, so there might be need to
        use another of the provided topo modes and provide the respective
        files locally.

        Parameters
        ----------
        p : str
            new search path for topography data
        """
        if not exists(p):
            raise FileExistsError("Input path does not exist")
        self.local_topo_path = p

    def change_topo_mode(self, new_mode=None, local_path=None):
        """Change the current mode for topography data access

        Parameters
        ----------
        new_mode : str
            new topo access mode, default to "srtm"
        local_path : str, optional
            if not None and valid, update local topo access

        """
        if new_mode is None:
            new_mode = "srtm"
        if local_path is not None:
            self.set_local_topo_path(local_path)
        self.topo_access_mode = new_mode

    def get_topo(self) -> TopoData:
        """Get current topo data"""
        if not isinstance(self.topo_data, TopoData):
            self.load_topo_data()
        return self.topo_data

    def load_topo_data(self):
        """Load topography data

        Note
        ----

        The loaded :class:`TopoData` object will also be set in all
        :class:`GeoPoint` objects belonging to this setup

        Parameters
        ----------
        topo_access_mode : str, optional
            topo dataset that is supposed to be used. If None, then
            :attr:`topo_access_mode` is used.

        Returns
        -------
        TopoData
            topographic data (is also assigned to :attr:`topo_data`).
        """
        if "ll" not in self.points:
            self.set_borders_from_points()
        self.topo_data = self.topo_access.get_data(self.ll.latitude,
                                                   self.ll.longitude,
                                                   self.tr.latitude,
                                                self.tr.longitude)
        for p in list(self.points.values()):
            p.set_topo_data(self.topo_data)

        return self.topo_data


    def has_point(self, name):
        """Checks if point with input name exists

        Parameters
        ----------
        name : str
            name of GeoPoint to be checked

        Returns
        -------
        bool
            Whether or not point exists
        """
        if name in self.points:
            return True
        return False

    def has_vector(self, name):
        """Checks if vector with input name exists

        Parameters
        ----------
        name : str
            name of vector

        Returns
        -------
        bool
            whether or not such a vector with input name exists
        """
        if name in self.vectors:
            return True
        return False

    def contains_coordinate(self, lat, lon):
        """
        Check if input coordinate is within domain of this GeoSetup

        Parameters
        ----------
        lat : float
            latitude of coordinate
        lon : float
            longitude of coordinate

        Returns
        -------
        bool

        """
        latok = self.lat_ll <= lat <= self.lat_tr
        lonok = self.lon_ll <= lon <= self.lon_tr
        return True if latok and lonok else False

    def add_geo_point(self, pt, assert_in_domain=True,
                      overwrite_existing=False) -> None:
        """Add a GeoPoint to this collection

        Parameters
        ----------
        pt : GeoPoint
            point to be added
        assert_in_domain : bool
            if True, a check is performed whether the input point is within
            domain or not (does not apply if input point is "ll" or "tr", that
            is, one of the points defining the domain itself). Defaults to
            True.
        overwrite_existing : bool
            if True and a point with the same name already exists in this
            setup, then the existing point will be overwritten with input
            point, else an exception is raises. Defaults to False.

        Raises
        ------
        OutOfDomain
            if paramerter `assert_in_domain` is True and input point is not within
            current domain.
        ValueError
            if parameter `overwrite_existing` is False and a point with the input n
            name already exists in this setup.

        Returns
        -------
        None
        """
        if not isinstance(pt, GeoPoint):
            raise ValueError(f'invalid input: {pt}. Need GeoPoint')
        if pt.name == 'll':
            self.ll = pt
        elif pt.name == 'tr':
            self.tr = pt
        elif pt.name in self.points and not overwrite_existing:
            raise ValueError(
                f'GeoPoint with name {pt.name} already exists in GeoSetup. '
            )
        elif self.borders_set and assert_in_domain and not \
                self.contains_coordinate(pt.latitude, pt.longitude):

            raise OutOfDomain(f'{pt} is not within domain of GeoSetup')

        else:
            self.points[pt.name] = pt
            if (isinstance(self.topo_data, TopoData) and
                    not isinstance(pt.topo_data, TopoData)):
                try:
                    pt.set_topo_data(self.topo_data)
                except OutOfDomain:
                    pass

    def add_geo_points(self, *pts, assert_in_domain=False) -> None:
        """Add multiple GeoPoints to the collection

        Parameters
        ----------
        *pts
            points to add
        assert_in_domain : bool
            if True, check assert that each point is within domain

        Returns
        -------
        None
        """
        for pt in pts:
            self.add_geo_point(pt, assert_in_domain)

    def add_geo_vector(self, vec, overwrite_existing=True) -> None:
        """Add :class:`GeoVector3D` to this collection

        Parameters
        ----------
        vec : GeoVector3D
            vector to be added
        overwrite_existing : bool
            If True and if vector with same name already exists, the former one

        Raises
        ------
        ValueError
            if input is not GeoVector3D

        """
        if not isinstance(vec, GeoVector3D):
            raise ValueError(f'invalid input: {vec}. Need GeoVector3D')
        if vec.name in self.vectors and not overwrite_existing:
            raise ValueError(f'Vector with name {vec.name} already exists.')
        self.vectors[vec.name] = vec

    def add_geo_vectors(self, *vecs) -> None:
        """Add multiple GeoVector3D objects to the collection

        Parameters
        ----------
        *vecs
            Instances of :class:`GeoVector3D` to be added
        """
        for vec in vecs:
            self.add_geo_vector(vec)

    def delete_geo_point(self, name) -> None:
        """Remove one of the geo points from the collection

        Parameters
        ----------
        name : str
            name of the point

        Raises
        ------
        ValueError
            if no point with with input name exists in this setup

        Returns
        -------
        None
        """
        if not name in self.points:
            raise ValueError(f'no such GeoPoint ({name}) in GeoSetup')
        del self.points[name]

    def delete_geo_vector(self, name):
        """Remove one of the vectors from the collection

        Parameters
        ----------
        name : str
            name of the vector

        Raises
        ------
        ValueError
            if no vector with with input name exists in this setup

        Returns
        -------
        None
        """
        if not name in self.vectors:
            raise ValueError(f'no such GeoVector3D ({name}) in GeoSetup')
        del self.vectors[name]

    def new_geo_point(self, *args, **kwargs) -> None:
        """Create new geo_point and add to collection

        Create GeoPoint using input args and kwargs and then parse that
        GeoPoint to :func:`add_geo_point`.

        Parameters
        ----------
        *args
            input args passed to init function of GeoPoint
        **kwargs
            input keyword args passed to init function of GeoPoint

        Returns
        -------
        None
        """
        self.add_geo_point(GeoPoint(*args, **kwargs))

    def _all_lats_lons(self) -> tuple:
        """Get list of all latitude and longitude coordinates of all points

        Returns
        ----
        ndarray
            latitude coordinates
        ndarray
            longitude coordinates
        """
        lats, lons = [], []
        for p in self.points.values():
            lats.append(p.latitude)
            lons.append(p.longitude)
        return (asarray(lats), asarray(lons))

    def set_borders_from_points(self, extend_km=1, to_square=True) -> None:
        """Set lower left (ll) and top right (tr) corners of domain

        The domain is inferred from all points associated with this
        setup.

        Parameters
        ----------
        extend_km : float
            extend range from the outermost points by this number in km
        to_square : bool
            extend the shorter base side to the size of the longer one (
            quadratic range)

        Raises
        ------
        AttributeError
            if no points are available in the setup.
        """
        lats, lons = self._all_lats_lons()
        if not len(lats) > 0:
            raise AttributeError('Cannot initiate range coordinates in empty '
                                 'GeoSetup, please add at least one point')

        lat_ll, lon_ll, lat_tr, lon_tr = (nanmin(lats), nanmin(lons),
                                          nanmax(lats), nanmax(lons))

        pll, ptr = GeoPoint(lat_ll, lon_ll, 0.0), GeoPoint(lat_tr, lon_tr, 0.0)

        if to_square:
            v = ptr - pll
            add = (abs(v.dx) - abs(v.dy)) / 2
            if add > 0:  # E/W extend (dx) is smaller than N/S extend (dy)
                pll = pll.offset(azimuth=180, dist_hor=add)
                ptr = ptr.offset(azimuth=0, dist_hor=add)
            else:
                pll = pll.offset(azimuth=270, dist_hor=-add)
                ptr = ptr.offset(azimuth=90, dist_hor=-add)

        ll = pll.offset(azimuth=-135,
                        dist_hor=float(extend_km),
                        name='ll')
        self.add_geo_point(ll, assert_in_domain=False,
                           overwrite_existing=True)

        tr = ptr.offset(azimuth=45,
                        dist_hor=float(extend_km),
                        name='tr')
        self.add_geo_point(tr, assert_in_domain=False,
                           overwrite_existing=True)

    def points_close(self, p, radius=None):
        """
        Find all GeoPoints within a certain radius around input point

        The search is performed against all points defined in this GeoSetup.

        Parameters
        ----------
        p : GeoPoint
            location for which the search is performed
        radius : float, optional
            radius (in km) specifying considered range around point (is set
            to 10% of the magnitude of this setup if unspecified)

        Returns
        -------
        list
            names of all points that are in vicinity around input point
        """
        if not isinstance(p, GeoPoint):
            raise ValueError('invalid input, need GeoPoint')
        if radius == None:
            radius = self.magnitude * .1
        names = []
        for pt in list(self.points.values()):
            if not pt is p and (pt - p).magnitude < radius:
                names.append(pt.name)
        print(f"Found {len(names):d} points within radius of {radius:.1f} km "
              f"of point {p.name}")
        return names

    @staticmethod
    def create_test_setup() -> 'GeoSetup':
        """Initiate example test data set

        Returns
        -------
        GeoSetup
            example setup
        """
        gs = GeoSetup()
        source = GeoPoint(latitude=37.751005, longitude=14.993435,
                          name="Etna", auto_topo_access=True)
        instrument = GeoPoint(latitude=37.765755, longitude=15.016696,
                              name="Observatory", auto_topo_access=True)
        gs.add_geo_points(source, instrument)
        gs.set_borders_from_points()
        plume = GeoVector3D(azimuth=83, dist_hor=gs.magnitude,
                            elevation=0, anchor=source, name="plume")
        view_dir = GeoVector3D(azimuth=160, dist_hor=gs.magnitude,
                               elevation=8, anchor=instrument, name="cfov")

        gs.add_geo_vectors(plume, view_dir)
        return gs

    def create_map(self, *args, **kwargs):  # pragma: no cover
        """Create a Basemap object for this regime"""
        if not BASEMAP_AVAILABLE:
            raise ImportError("Cannot create map: "
                              "Basemap library is not available")

        if not isinstance(self.topo_data, TopoData):
            self.load_topo_data()
        if not "projection" in kwargs and self.magnitude < 150:
            kwargs["projection"] = "lcc"
        if "llcrnrlon" not in kwargs:
            kwargs["llcrnrlat"] = self.lat_ll
            kwargs["llcrnrlon"] = self.lon_ll
            kwargs["urcrnrlat"] = self.lat_tr
            kwargs["urcrnrlon"] = self.lon_tr
            kwargs["lat_0"], kwargs["lon_0"] = self.center_coordinates

        m = Map(*args, **kwargs)
        m.set_topo_data(self.topo_data)
        return m

    def plot_2d(self, draw_all_points=True, draw_all_vectors=True,
                draw_topo=True, draw_coastline=True, draw_mapscale=True,
                draw_legend=True, *args, **kwargs):  # pragma: no cover
        """
        To be updated
        """
        if not BASEMAP_AVAILABLE:
            raise ImportError("Cannot create overview map: Basemap module "
                              "is not available")
        if "ax" not in kwargs:
            from matplotlib.pyplot import figure
            fig = figure(figsize=(10, 8))
            ax = fig.add_axes([0.12, 0.15, 0.8, 0.8])
            kwargs["ax"] = ax
        m = self.create_map(*args, **kwargs)
        if draw_coastline:
            try:
                m.drawcoastlines()
            except ValueError:
                pass  # see https://github.com/jgliss/geonum/issues/5
        if draw_topo:
            m.draw_topo(insert_colorbar=True)
            m.draw_topo_contour()
        m.draw_coordinates()
        if draw_mapscale:
            m.draw_mapscale_auto()
        p_close_count = 0
        if draw_all_points:
            dist = self.magnitude * .05
            for pt in list(self.points.values()):
                if not any([pt.name == x for x in ["ll", "tr"]]):
                    m.draw_geo_point_2d(pt)
                    ang = -45
                    num_close = len(self.points_close(pt))
                    if num_close > 0:
                        step = 360. / (4 * num_close)
                        ang = ang - step * p_close_count
                        p_close_count += 1
                    m.write_point_name_2d(pt, dist, ang)

        # create some color indices for colormap
        nums = [int(255.0 / k) for k in range(1, len(self.vectors) + 3)]
        if draw_all_vectors:
            for i, vec in enumerate(self.vectors.values()):
                m.draw_geo_vector_2d(vec,
                                     ls="-",
                                     c=self.cmap_vecs(nums[i]),
                                     label=vec.name)
        if draw_legend:
            try:
                m.legend()
            except Exception as e:
                warn(f"Failed to draw legend in GeoSetup...: {e}")

        return m

    def plot_3d(self, draw_all_points=True, draw_all_vectors=True,
                cmap_topo="Oranges", contour_color="#708090",
                contour_lw=0.2, contour_antialiased=True,
                *args, **kwargs):  # pragma: no cover
        """Create a 3D overview map of the current setup

        Parameters
        ----------
        draw_all_points : bool
            if True, all current GeoPoint objects are plotted, defaults to
            True
        draw_all_vectors : bool
            if True, all current GeoVector3D objects are plotted, defaults to
            True
        cmap : str
            string ID of the colormap used to plot the topographic data,
            defaults to "Oranges"
        contour_color : str
            string specifying color of contour lines colors of contour lines
            (default: "#708090")
        contour_lw :
            width of drawn contour lines, defaults to 0.5, use 0 if you do not
            want contour lines inserted
        contour_antialiased : bool
            apply antialiasing to surface plot of topography, defaults to False
        *args :
            additional non-keyword parameters (passed to `basemap
            <http://matplotlib.org/basemap/api/basemap_api.html#mpl
            _toolkits.basemap.Basemap>`_)
        **kwargs :
            additional keyword parameters (passed to `basemap
            <http://matplotlib.org/basemap/api/basemap_api.html#mpl
            _toolkits.basemap.Basemap>`_)

        Returns
        -------
        Map
            plotted 3D basemap
        """
        if not BASEMAP_AVAILABLE:
            raise ImportError("Cannot create overview map: Basemap module "
                              "is not available.")
        m = self.create_map(*args, **kwargs)
        m.draw_topo_3d(cmap=cmap_topo, contour_color=contour_color,
                       contour_lw=contour_lw,
                       contour_antialiased=contour_antialiased)
        if draw_all_points:
            zr = self.topo_data.alt_range * 0.05
            alts = []
            for name, pt in self.points.items():
                if not any([name == x for x in ["ll", "tr"]]):
                    try:
                        add_alt = 0  # in m
                        for alt in alts:
                            if abs(alt - pt.altitude) < zr:
                                add_alt = 3 * zr
                                print("Add " + str(add_alt))

                        pt.plot_3d(m, add_name=True, dz_text=zr + add_alt)
                        alts.append(pt.altitude)

                    except Exception as e:
                        warn("Point %s could not be drawn: %s"
                             % (pt.name, repr(e)))

        if draw_all_vectors:
            nums = [int(255.0 / k) for k in range(1, len(self.vectors) + 3)]
            for i, vec in enumerate(self.vectors.values()):
                try:
                    m.draw_geo_vector_3d(vec, label=vec.name,
                                         c=self.cmap_vecs(nums[i]),
                                         ls="-",
                                         **kwargs)
                except Exception as e:
                    warn("Vector %s could not be drawn: %s"
                         % (vec.name, repr(e)))

        return m



# -*- coding: utf-8 -*-
#
# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Access and handling of topographic data
"""
import pytest
import numpy as np
import numpy.testing as npt
from geonum.conftest import does_not_raise_exception
from geonum.exceptions import TopoAccessError
from geonum import LOCAL_TOPO_DIR
from geonum import topoaccessbase as mod

@pytest.fixture(scope='module')
def etopo_acc():
    return mod.Etopo1Access(None, None, False, False)

@pytest.mark.parametrize('local_path, file_name, check_access,search_database,'
                         'raises', [
    (None, None, False, True,does_not_raise_exception()),
    (None, None, False, False,does_not_raise_exception()),
    (None, None, True, False,pytest.raises(TopoAccessError)),

])
def test_Etopo1Access___init__(tmpdir, local_path, file_name, check_access,
                               search_database,raises):

    with raises:
        val = mod.Etopo1Access(local_path=local_path,
                               file_name=file_name,
                               check_access=check_access,
                               search_database=search_database)
        if file_name is None:
            assert val._file_name == "ETOPO1_Ice_g_gmt4.grd"
        else:
            assert val._file_name == file_name

        if local_path is None:
            assert val._local_path == LOCAL_TOPO_DIR
        else:
            assert val._local_path == local_path

def test_Etopo1Access_check_access(etopo_acc):
    assert not etopo_acc.check_access()

def test_Etopo1Access_check_access_MOCK_GET_DATA(
        monkeypatch):
    def mock_get_data(self, lat0, lon0, lat1=None, lon1=None):
        return 42

    monkeypatch.setattr(mod.Etopo1Access, 'get_data', mock_get_data)

    acc = mod.Etopo1Access(local_path=None,
              file_name=None,
              check_access=False,
              search_database=False)
    assert acc.check_access()

@pytest.mark.parametrize('lat0,lon0,lat1,lon1,res', [
    (6,1,None,None,
     (np.asarray([6]), np.asarray([1]), [1,1], [1,1])),
    (8,4,-1,-1,
     (np.asarray([5,6,7]), np.asarray([0,1,2,3]), [0, 2], [0, 3]))
])
def test__init_lons_lats(etopo_acc,lat0,lon0,lat1,lon1,res):
    lats_all = np.asarray([5,6,7])
    lons_all=np.asarray([0,1,2,3])
    result = etopo_acc._init_lons_lats(
        lats_all, lons_all,lat0,lon0,lat1,lon1)
    for i, item in enumerate(res):
        npt.assert_array_equal(result[i], item)

def test_init_Etopo1_invalid_file(tmpdir):
    file_name = 'invalid.grd'
    path = tmpdir / file_name
    with open(path, 'w') as f:
        f.write('bla')

    assert path.exists()
    with pytest.raises(TopoAccessError):
        mod.Etopo1Access(local_path=tmpdir,
                               file_name=file_name,
                               check_access=True,
                               search_database=False)








# -*- coding: utf-8 -*-
"""Test environment for base.py module."""
import pytest
import numpy.testing as npt
from geonum.conftest import skip_srtm, skip_netcdf4
import geonum.topodataaccess as mod

@pytest.fixture
def acc():
    return mod.TopoDataAccess()

@skip_netcdf4
def test_etopo1_init():
    acc = mod.Etopo1Access(check_access=False, search_database=False)
    from geonum import LOCAL_TOPO_DIR
    assert LOCAL_TOPO_DIR == acc.local_path
    assert acc.file_name == 'ETOPO1_Ice_g_gmt4.grd'

@skip_srtm
def test_srtm():
    acc = mod.SRTMAccess()
    d = acc.get_data(acc._TESTLAT, acc._TESTLON)
    npt.assert_equal(d.data.shape, (1, 1))
    assert d.data[0][0] == 861

@skip_srtm
@pytest.mark.parametrize('lat0, lon0, lat1, lon1, dsh, dmin, dmax, dmean', [
    (60.052, 7.414, None, None, (2, 2), 1153., 1183., 1167.5),
    (-18.55, -69.2, -18.35, -69.0, (241, 121), 4084., 6057., 4690.5),

    ])
def test_srtm_access(lat0, lon0, lat1, lon1, dsh, dmin, dmax, dmean):
    acc = mod.TopoDataAccess(mode='srtm')
    data = acc.get_data(lat0, lon0, lat1, lon1)
    npt.assert_array_equal(data.shape, dsh)
    npt.assert_allclose([data.min, data.max, data.mean()],
                                  [dmin, dmax, dmean], atol=0.1)

def test_topoaccess_invalid_mode():
    all_ok = True
    try:
        mod.TopoDataAccess('bla')
    except mod.InvalidTopoMode:
        pass
    except Exception:
        all_ok = False

    acc = mod.TopoDataAccess()
    try:
        acc.get_data(0,0, mode='bla')
    except mod.InvalidTopoMode:
        pass
    except Exception:
        all_ok = False
    assert all_ok

def test_modes(acc):
    assert acc.modes == ['srtm', 'etopo1']

if __name__ == "__main__":
    import sys
    pytest.main(sys.argv)


# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import matplotlib.pyplot as plt
import numpy as np
from scipy.ndimage import map_coordinates


class LineOnGrid(object):
    """A class representing a line on a discrete grid

    Attributes
    ----------
    start : list
        start location [x0, y0]
    stop : list
        stop location [x1, y1]

    Parameters
    ----------
    x0 : int
        start x coordinate
    y0 : int
        start y coordinate
    x1 : int
        end x coordinate
    y1 : int
        end y coordinate
    name : int
        string for identification (optional)

    Example
    -------

    >>> import numpy as np
    >>> from geonum import LineOnGrid
    >>> data = np.ones((10,15)) * np.arange(15)
    >>> line = LineOnGrid(2,2,5,5)
    >>> profile = line.get_line_profile(data)
    >>> print(profile)
    [2.         2.74392912 3.50304939 4.24920976 5.        ]

    """
    def __init__(self, x0, y0, x1, y1, name=None):
        if name is None:
            name = 'undefined'
        self.name = name
        self.start = [x0, y0]
        self.stop = [x1, y1]

        self._profile_coords = None

        self._det_normal_vec()

    @property
    def profile_coords(self) -> 'numpy.ndarray':
        """(y,x) profile coordinates"""
        if self._profile_coords is None:
            self._profile_coords = self._prepare_profile_coordinates()
        return self._profile_coords


    def _prepare_profile_coordinates(self):
        """Prepare the evaluation coordinates as stack"""
        length = self.length
        x0, y0 = self.start
        x1, y1 = self.stop
        x = np.linspace(x0, x1, length + 1)
        y = np.linspace(y0, y1, length + 1)
        return np.vstack((y, x))

    @property
    def normal_vector(self) -> tuple:
        """Normal vector of line (x, y)"""
        return self._det_normal_vec()

    @property
    def length(self) -> int:
        """Determine the length in units of grid indices

        Note
        ----
        The length is rounded to integer precision.
        """
        del_x, del_y = self._delx_dely()
        return int(round(np.hypot(del_x, del_y)))

    def get_line_profile(self, array, **kwargs):
        """Retrieve line profile of data on a 2D array

        Uses method :func:`scipy.ndimage.map_coordinates`

        Note
        ----
        The spline interpolation will fail if there are NaNs in the input data.
        In this case, try using order=1 as additional input argument or use
        prefilter=False. For details see `here <https://docs.scipy.org/doc/
        scipy-0.14.0/reference/generated/scipy.ndimage.interpolation.
        map_coordinates.html>`__

        Parameters
        ----------
        array : numpy.ndarray
            2D data array
        **kwargs
            additional keyword args that are passed to
            :func:`scipy.ndimage.map_coordinates`

        Returns
        -------
        numpy.ndarray
            1D array containing the values along the line profile coordinates
        """
        return map_coordinates(array, self.profile_coords, **kwargs)

    def plot_line_on_grid(self, array, ax=None, **kwargs):
        """Plot this line on an input array using imshow

        Parameters
        ----------
        array : numpy.ndarray
            the data array to which this line is mapped
        ax : matplotlib.axes.Axes, optional
            axes object
        **kwargs
            additional keyword arguments for imshow

        Returns
        -------
        matplotlib.axes.Axes
            axes instance used for plotting

        """
        if ax is None:
            fig, ax = plt.subplots(1, 1)
        else:
            fig = ax.figure

        imshow_args = dict(cmap="gray", interpolation='none', origin="upper")
        imshow_args.update(kwargs)
        im = ax.imshow(array, **imshow_args)
        fig.colorbar(im, ax=ax, shrink=0.9)
        ax.plot([self.start[0], self.stop[0]],
                [self.start[1], self.stop[1]],
                'co-')

        ax.set_xlim([0, array.shape[1] - 1])
        ax.set_ylim([array.shape[0] - 1, 0])
        return ax

    def plot_line_profile(self, array, ax=None):
        """Plot profile of line in input data array

        Parameters
        ----------
        array : numpy.ndarray
            the data array to which this line is mapped
        ax : matplotlib.axes.Axes, optional
            axes object

        Returns
        -------
        matplotlib.axes.Axes
            axes instance used for plotting
        """
        if ax is None:
            fig, ax = plt.subplots(1, 1)
        p = self.get_line_profile(array)
        ax.set_xlim([0, self.length])
        ax.grid()
        ax.plot(p)
        ax.set_title("Line profile")
        return ax

    def plot(self, array):
        """High level plot function, calls:

            1. :func:`plot_line_on_grid`
            2. :func:`plot_line_profile`

        and puts them next to each other in a subplot

        Parameters
        ----------
        array : numpy.ndarray
            2D data array to which this line is mapped

        Returns
        -------
        matplotlib.figure.Figure
            figure instance containing the plots.
        """
        fig, axes = plt.subplots(2, 1)

        self.plot_line_on_grid(array, ax=axes[0])
        self.plot_line_profile(array, ax=axes[1])
        plt.tight_layout()
        return fig

    def _delx_dely(self):
        """Returns length of x and y range"""
        return (abs(self.stop[0] - self.start[0]),
                abs(self.stop[1] - self.start[1]))

    def _det_normal_vec(self):
        """Determines the normal vector (orientation) of the line

        Returns
        -------
        tuple
            2-element tuple containing (x, y) elements of normal vector
        """
        delx, dely = self._delx_dely()
        try:
            a = np.arctan(float(dely) / float(delx))
        except ZeroDivisionError:
            a = np.pi / 2
        return (np.sin(a), np.cos(a))

    def __str__(self):
        """String representation"""
        s = (
            f"LineOnGrid {self.name}. Start (x,y): {self.start}. Stop (x,"
            f"y): {self.stop}")
        return s



# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from LatLon23 import LatLon

from geonum.exceptions import OutOfDomain
from geonum.topodata import TopoData
from geonum.topodataaccess import TopoDataAccess


class GeoPoint(LatLon):
    """The Geopoint object represents a location in the atmosphere

    This object is in 3D and includes altitude information.

    Attributes
    ----------
    altitude : float
        altitude in m
    altitude_err : float
        uncertainty in altitude
    local_topo_path : str
        directory that is used to search for local topography if topodata is
        requested. Irrelevant for default topo

    Parameters
    ----------
    latitude : float, optional
        latitude of point (decimal degrees), defaults to 0.
    longitude : float, optional
        longitude of point (decimal degrees), defaults to 0.
    altitude : float, optional
        elevation of point in m, defaults to 0.
    name : str, optional
        name (ID) of this point, defaults to undefined
    topo_access_mode : str, optional
        string specifying the current access mode for topographic data (in v1,
        choose between srtm or etopo1), defaults to None, in which case default
        of :class:`TopoDataAccess` is used if topographic data is accessed via
        this point (e.g. if `auto_topo_access` is True). As of v1.4.X, default
        access mode is srtm, which uses SRTM database, which does not have
        full global coverage!
    topo_path : str, optional
        local path where local topography data is stored (at the moment only
        relevant for topo_mode etopo1), defaults to None.
    topo_data : TopoData, optional
        existing topographic dataset that can be assigned to this point (can
        save time, e.g. for altitude access)
    auto_topo_access : bool
        if True, try set :attr:`altitude`and :attr:`altitude_err` based on
        local topography (note that defa)
    """
    _ALTERR_DEFAULT = 1e9

    def __init__(self, latitude=0, longitude=0, altitude=None, name=None,
                 topo_access_mode=None, topo_path=None, topo_data=None,
                 auto_topo_access=False):
        if name is None:
            name = 'undefined'
        LatLon.__init__(self, float(latitude), float(longitude), name)
        if altitude is None:
            altitude = 0
        elif auto_topo_access:  # altitude is not None
            raise ValueError(
                'either provide altitude or deactivate auto_topo_access'
            )
        self.altitude = altitude  # altitude in m
        self.altitude_err = self._ALTERR_DEFAULT

        self.topo_access_mode = topo_access_mode
        self.local_topo_path = topo_path

        self.topo_data = None

        if topo_data is not None:
            self.set_topo_data(topo_data)

        if auto_topo_access:
            self.set_topo_altitude()

    @property
    def longitude(self) -> float:
        """Longitude coordinate in decimal degrees"""
        return self.lon.decimal_degree

    @property
    def latitude(self) -> float:
        """Latitude coordinate in decimal degrees"""
        return self.lat.decimal_degree

    @property
    def _topo_access(self) -> TopoDataAccess:
        """Topography data access class"""
        return TopoDataAccess(mode=self.topo_access_mode,
                              local_path=self.local_topo_path)

    def offset(self, azimuth, dist_hor, dist_vert=0.0, ellipse=None,
               **kwargs) -> 'GeoPoint':
        """Returns new GeoPoint at offset position

        Parameters
        -----------
        azimuth : float
            azimuth direction angle (in decimal degrees, e.g. 90 for E
            direction, -90 or 270 for west)
        dist_hor : float
            horizontal offset to this point in km
        dist_vert : float
            vertical offset to this point in m (positive if point is higher,
            negative, if it is lower)
        ellipse : float
            geodetic system (ellipsoid), default is "WGS84", i.e.
            `World Geodetic System <https://confluence.qps.nl/pages/view page.
            action?pageId=29855173>`__
        **kwargs :
            additional keyword arguments passed to init of new
            :class:`GeoPoint` object

        Returns
        -------
        GeoPoint
            new location at offset position
        """
        if ellipse is None:
            ellipse = "WGS84"
        p = LatLon(self.lat.decimal_degree, self.lon.decimal_degree)
        p1 = p.offset(azimuth, dist_hor, ellipse)
        return GeoPoint(p1.lat.decimal_degree, p1.lon.decimal_degree,
                        self.altitude + dist_vert, **kwargs)

    def set_topo_data(self, topo_data):
        """Assign topographic data to this point

        If input is valid, the topographic data set will be stored in
        class attribute topo_data.

        Parameters
        ----------
        topo_data : TopoData
            topography dataset
        """
        if not isinstance(topo_data, TopoData):
            raise ValueError('need instance of geonum.TopoData')
        elif not topo_data.includes_coordinate(self.latitude, self.longitude):
            raise OutOfDomain(
                f'Cannot assign input TopoData {topo_data} to {self}. '
                f'The location of {self} are outside the borders of TopoData'
            )
        self.topo_data = topo_data

    def range_borders(self, *points, extend_fac=0.1):
        """Get geographical borders (lower left, top right) of this point

        Additional points can be included as non keyword arguments. The
        range borders are then determined under consideration of all
        specified points. If no additional points are specified,
        the range corners will be set corresponding to 1km diagonal
        distance to this point.

        Parameters
        ----------
        *points
            additional :class:`GeoPoint` objects to be  considered
        extend_fac : float
            domain extension factor wrt to lat / lon span covered by input
            points, defaults to 0.1 (ca 10% extension)

        Returns
        -------
        GeoPoint
            lower left corner of regime
        GeoPoint
            top right corner of regime

        """
        lats, lons = [self.latitude], [self.longitude]
        # retrieve all latitudes and longitudes
        for p in points:
            if isinstance(p, GeoPoint):
                lats.append(p.latitude)
                lons.append(p.longitude)
        lats, lons = np.asarray(lats), np.asarray(lons)

        (lat_ll, lon_ll,
         lat_tr, lon_tr) = (np.nanmin(lats), np.nanmin(lons),
                            np.nanmax(lats), np.nanmax(lons))
        pll, ptr = GeoPoint(lat_ll, lon_ll, 0.0), GeoPoint(lat_tr, lon_tr, 0.0)
        extend_km = (pll - ptr).magnitude * extend_fac
        if extend_km == 0:
            extend_km = 1
        ll = pll.offset(azimuth=-135, dist_hor=float(extend_km), name="ll")
        tr = ptr.offset(azimuth=45, dist_hor=float(extend_km), name="tr")
        return (ll, tr)

    def get_topo_data(self, geo_point=None, azimuth=None, dist_hor=None,
                      lon1=None, lat1=None):
        """Retrieve topographic data grid spanned by this and another point

        The second coordinate can be specified in several different ways
        dependent on which of the input arguments are used. First checks if the
        currently loaded topo data (``self.topo_data``) covers the range
        spanned by the 2 points. If this is the case, the loaded data will be
        used and nothing reloaded. If not, data will attempted to be loaded
        from the current :class:`TopoDataAccess` object within the lon lat
        range covered by the 2 points (set using :func:`range_borders`)

        Note
        ----
        The following input combinations work (and are preferentially
        processed in the specified order if multiple input is given):

            1. specify endpoint using geo_point
            #. specify endpoint using azimuth and dist_hor
            #. specify endpoint using lon1 and lat1

        Parameters
        ----------
        geo_point : GeoPoint
            another geo_point,topo data will be retrieved between this point
            and destination point
        azimuth : float
            azimuth angle of heading in decimal degrees (to be used with
            dist_hor)
        dist_hor : float
            horizontal distance from this point in km (to be used with azimuth)
        lon1 : float
            longitude of destination point, topo data will be retrieved between
            this point and destination point (to be used with lat1)
        lat1 : float
            latitude of destination point, topo data will be retrieved between
            this point and destination point (to be used with lon1)

        Returns
        -------
        TopoData
            topographic data object.
        GeoPoint
            the second coordinate which deterimes the domain wrt to this
            location.
        """
        pf = self
        if isinstance(geo_point, GeoPoint):
            pf = geo_point
        elif all(isinstance(x, (int, float)) for x in [dist_hor, azimuth]):
            pf = self.offset(azimuth, dist_hor)
        elif all(isinstance(x, (int, float)) for x in [lon1, lat1]):
            pf = GeoPoint(lat1, lon1)

        ll, tr = self.range_borders(self, pf)
        lat0, lon0 = ll.latitude, ll.longitude
        lat1, lon1 = tr.latitude, tr.longitude
        if not (self.check_topo(lat1, lon1) and self.check_topo(lat0, lon0)):
            self._load_topo_data(lat0, lon0, lat1, lon1)
        return self.topo_data, pf

    def check_topo(self, lat1=None, lon1=None):
        """Check if topography is available between this point and another

        Parameters
        ----------
        lat1 : float
            latitude of end point, if None, only the coordinates of this
            object will be considered
        lon1 : float
            longitude of end point, if None, only the coordinates of this
            object will be considered

        Returns
        -------
        bool
            True if topo data is loaded, False if not
        """
        if any([x == None for x in [lat1, lon1]]):
            lat1, lon1 = self.latitude, self.longitude
        if self.topo_data is None:
            return False
        if not (self.topo_data.includes_coordinate(self.latitude,
                                                   self.longitude)
                and self.topo_data.includes_coordinate(lat1, lon1)):
            return False
        return True

    def get_elevation_profile(self, geo_point=None, azimuth=None,
                              dist_hor=None, lon1=None, lat1=None,
                              resolution=5., **mapping_opts):
        """Estimates the elevation profile for a given viewing direction

        The profile is retrieved up to a given distance from the coordinate of
        this point. For input possibilities see docs of :func:`get_topo_data`.

        Note
        ----
        The following input combinations work (and are preferentially processed
        in the specified list order if multiple input is given):

            1. specify endpoint using geo_point
            #. specify endpoint using azimuth and dist_hor
            #. specify endpoint using lon1 and lat1

        Parameters
        ----------
        geo_point : GeoPoint
            end location of elevation profile (if this is provided, the
            following other input options to determine the end point will be
            ignored).
        azimuth : float
            azimuth angle (direction) of elevation profile in decimal degrees
            (to be used with input arg dist_hor)
        dist_hor : float
            horizontal distance from this point in km (to be used with azimuth)
        lon1 : float
            longitude of destination point, topo data will
            be retrieved between this point and destination point (to be
            used with lat1)
        lat1 : float
            latitude of destination point, topo data will be
            retrieved between this point and destination point
            (to be used with lon1)
        resolution : float
            desired topo grid resolution in m.
            1D interpolation of the elevation profile is performed if
            applicable.

        Returns
        -------
        ElevationProfile
            the profile object
        """
        from geonum.elevationprofile import ElevationProfile
        topo, pf = self.get_topo_data(geo_point, azimuth, dist_hor, lon1, lat1)
        if pf == self:
            raise ValueError('please specify endpoint location')
        return ElevationProfile(observer=self,
                                endpoint=pf,
                                topo_data=topo,
                                calc_on_init=True,
                                resolution=resolution,
                                **mapping_opts)

    def set_topo_altitude(self):
        """Set altitude using topographic terrain height at lat / lon position

        The estimation is done by retrieving a 2x2 grid of topography
        data enclosing this point. The altitude value is estimated using the
        topo data tile closest to the coordinates of this point. The
        uncertainty is estimated conservatively using min/max difference of
        data in the 2x2 grid.

        Returns
        -------
        float
            altitude
        float
            uncertainty in altitude value
        """
        if not isinstance(self.topo_data, TopoData):
            data = self._topo_access.get_data(self.latitude,
                                              self.longitude)
            # store for later usage
            self.set_topo_data(data)
        else:
            data = self.topo_data
        z = data(self.latitude, self.longitude)
        z_err = (data.data.max() - data.data.min())
        self.altitude, self.altitude_err = z, z_err
        return (z, z_err)

    def plot_2d(self, map, add_name=False, dist_text=0.5, angle_text=-45,
                **kwargs):  # pragma: no cover
        """Plot this point into existing 2D basemap

        Parameters
        ----------
        map : basemap
            Basemap object (drawn in an Axes3D object)
        add_name : bool
            add the name of this GeoPoint in the map
        dist_text : float
            distance of text annotation from point
        angle_text : float
            angle of text displacement
        **kwargs
            additional keyword arguments passed to matplotlib plot function
        """
        if not "marker" in kwargs:
            kwargs["marker"] = "^"
        if not any([x in kwargs for x in ["c", "color"]]):
            kwargs["c"] = "lime"
        map.draw_geo_point_2d(self, **kwargs)
        if add_name:
            map.write_point_name_2d(self, dist_text, angle_text)

    def plot_3d(self, map, add_name=False, dz_text=0.0,
                **kwargs):  # pragma: no cover
        """
        Plot this point into existing 3D basemap

        map : basemap
            Basemap object (drawn in an Axes3D object)
        add_name : bool
            add the name of this GeoPoint in the map
        dz_text : float
            altitude offset of text (to point location in map)
        **kwargs
            additional keyword arguments passed to matplotlib plot function.
        """
        if not "marker" in kwargs:
            kwargs["marker"] = "^"
        if not any([x in kwargs for x in ["c", "color"]]):
            kwargs["c"] = "lime"
        map.draw_geo_point_3d(self, **kwargs)
        # x0, y0 = list(map(self.lon.decimal_degree, self.lat.decimal_degree))
        if add_name and self.name is not None:
            try:
                fs = kwargs["fontSize"]
            except Exception:
                fs = 12

            zt = self.altitude + dz_text
            map.draw_text_3d(self.longitude, self.latitude, zt, self.name,
                             color="k", fontsize=fs)

    def _load_topo_data(self, lat0, lon0, lat1, lon1) -> TopoData:
        """Load topo data

        Parameters
        ----------
        lon0 : float
            start longitude
        lat0 : float
            start latitude
        lon1 : float
            stop longitude
        lat1 : float
            stop latitude

        Return
        ------
        TopoData
            loaded topographic data
        """
        self.topo_data = self._topo_access.get_data(lat0, lon0, lat1, lon1)
        return self.topo_data

    def _sub_geo_vector_2d(self, other):
        """Subtract another geo vector

        Called when subtracting a GeoVector object from self (adapted from
        `LatLon` object,  only return type was changed from LatLon
        object to GeoPoint object)

        Parameters
        ----------
        other : GeoVector
            vector to be subtracted from this location

        Returns
        -------
        GeoPoint
            endpoint location
        """
        heading, distance = other()
        heading = (heading + 180) % 360  # Flip heading
        p = GeoPoint(self.lat, self.lon, self.altitude)

        return p.offset(heading, distance)

    def _add_geo_vector_2d(self, other):
        """Add another geo vector

        Add a geo vector (adapted from `LatLon` object,  only return
        type was changed from LatLon object to GeoPoint object)

        Parameters
        ----------
        other : GeoVector
            vector to be added to this location

        Returns
        -------
        GeoPoint
            endpoint location

        """
        azimuth, dist_hor = other()
        p = GeoPoint(self.lat, self.lon, self.altitude)
        return p.offset(azimuth, dist_hor, 0.0)

    def _sub_geo_vector_3d(self, other):
        """Subtract a GeoVector3D object from self

        Parameters
        ----------
        other : GeoVector3D
            vector to be subtracted from this location

        Returns
        -------
        GeoPoint
            endpoint location
        """
        azimuth, dist_hor, dz = other()
        azimuth = (azimuth + 180) % 360  # Flip heading
        p = GeoPoint(self.lat, self.lon, self.altitude)  # Copy current position
        return p.offset(azimuth, dist_hor, -dz)  # Offset position by GeoVector

    def _add_geo_vector_3d(self, other):
        """Add a 3d geo vector object

        Parameters
        ----------
        other : GeoVector3D
            vector to be added to this location

        Returns
        -------
        GeoPoint
            endpoint location
        """
        azimuth, dist_hor, dz = other()
        p = GeoPoint(self.lat, self.lon, self.altitude)
        return p.offset(azimuth, dist_hor, dz)

    def _sub_latlon(self, other):
        """Called when subtracting a LatLon object from this point"""
        from geonum import GeoVector3D
        inv = self._pyproj_inv(other)
        heading = inv['heading_reverse']
        distance = inv['distance']
        on = other.name
        if on is None:
            on = 'undefined'
        name = f'{on}->{self.name}'
        return GeoVector3D(dz=0.0, azimuth=heading, dist_hor=distance,
                           anchor=other, name=name)

    def _sub_geo_point(self, other):
        """Called when subtracting a LatLon object from self"""
        from geonum import GeoVector3D
        inv = self._pyproj_inv(other)
        heading = inv['heading_reverse']
        distance = inv['distance']
        name = str(other.name + "->" + self.name)
        dz = self.altitude - other.altitude
        return GeoVector3D(dz=dz, azimuth=heading, dist_hor=distance,
                           anchor=other, name=name)

    def __eq__(self, other):
        """Checks equality of this point with another

        Parameters
        ----------
        other : GeoPoint
            other location

        Returns
        -------
        bool
            True if other is equal, else False
        """
        if (self.latitude == other.latitude and
                self.longitude == other.longitude and
                self.altitude == other.altitude):
            return True
        return False

    def __add__(self, other):
        """Add a geo vector or geo point to this point"""
        object_operator = {'GeoVector': self._add_geo_vector_2d,
                           'GeoVector3D': self._add_geo_vector_3d}
        try:
            tp = other.type()
            if not tp in object_operator.keys():
                raise AttributeError
        except AttributeError:
            raise ValueError(f'invalid input type {type(other)}, choose from '
                             f'{list(object_operator)}')
        return object_operator[other.type()](other)

    def __sub__(self, other):
        """Subtraction"""
        object_operator = {'GeoVector': self._sub_geo_vector_2d,
                           'GeoVector3D': self._sub_geo_vector_3d,
                           'LatLon': self._sub_latlon,
                           'GeoPoint': self._sub_geo_point}
        try:
            tp = other.type()
            if not tp in object_operator.keys():
                raise AttributeError
        except AttributeError:
            raise ValueError(f'invalid input type {type(other)}, choose from '
                             f'{list(object_operator)}')
        return object_operator[other.type()](other)

    def __str__(self):
        """String formatting"""
        return (
            f"GeoPoint {self.name}\nLat: {self.latitude},  Lon:"
            f" {self.longitude}, Alt: {self.altitude} m\n")

    def __repr__(self):
        """Obj. representation"""
        return (
            f"Lat: {self.latitude}, Lon: {self.longitude}, Alt:"
            f" {self.altitude} m")

    def type(self):
        """Object type identifier

        Returns
        -------
        str
            value='GeoPoint'
        """
        return 'GeoPoint'

    @staticmethod
    def from_LatLon(coord):
        if not isinstance(coord, LatLon):
            raise ValueError('Need LatLon...')
        return GeoPoint(latitude=coord.lon.decimal_degree,
                        longitude=coord.lat.decimal_degree,
                        altitude=0,
                        auto_topo_access=False)



import cartopy
import numpy as np
import pytest

from geonum import plot_helpers as mod
from geonum.conftest import does_not_raise_exception


@pytest.mark.parametrize('projection', [
    None, cartopy.crs.TransverseMercator(), cartopy.crs.PlateCarree(),
    cartopy.crs.RotatedPole()
])
def test_init_figure_with_geoaxes(projection):
    ax = mod.init_figure_with_geoaxes()
    assert isinstance(ax, cartopy.mpl.geoaxes.GeoAxes)


@pytest.mark.parametrize('ax,xticks,yticks,xres,yres,raises', [
    (mod.init_figure_with_geoaxes(), None, None,
     [-180, -120, -60, 0., 60., 120., 180.],
     [-90., -60., -30., 0., 30., 60., 90.],
     does_not_raise_exception()),
    (mod.init_figure_with_geoaxes(xlim=(15.05, 15.08), ylim=(-0.023, 0.012)),
     None, None,
     [15.05, 15.056, 15.062, 15.068, 15.074, 15.08],
     [-0.023, -0.016, -0.009, -0.002, 0.005, 0.012],
     does_not_raise_exception()),
])
def test_set_map_ticks(ax, xticks, yticks, xres, yres, raises):
    with raises:
        ax = mod.set_map_ticks(ax, xticks, yticks)
        assert ax is ax
        xt = ax.get_xticks()
        yt = ax.get_yticks()
        np.testing.assert_allclose(xt, xres)
        np.testing.assert_allclose(yt, yres)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Feb 22 12:53:56 2021

@author: jonasg
"""
import os
import pytest
import srtm

def test_filehandler_access():
    fh = srtm.utils.FileHandler()

def test_filehandler_local_cache_dir():
    fh = srtm.utils.FileHandler()
    assert 'local_cache_dir' in fh.__dict__
    assert not fh.exists('nonexistingfile.hgt')
    assert os.path.exists(fh.local_cache_dir)

def test_get_data():
    acc = srtm.get_data()
    assert isinstance(acc, srtm.data.GeoElevationData)

if __name__ == "__main__":
    import sys
    pytest.main(sys.argv)



# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
from ._init_helpers import _check_requirements, _init_local_topodir, \
    _init_dir_and_version

LOCAL_TOPO_DIR, TOPO_INFO_FILE = _init_local_topodir()

(BASEMAP_AVAILABLE,
 CV2_AVAILABLE,
 NETCDF_AVAILABLE) = _check_requirements()

__dir__, __version__ = _init_dir_and_version()

from . import exceptions
from . import helpers
from . import atmosphere
from . import plot_helpers

from .topodata import TopoData
from .topodataaccess import TopoDataAccess

from .geopoint import GeoPoint
from .geovector3d import GeoVector3D
from .geosetup import GeoSetup
from .elevationprofile import ElevationProfile
from .lineongrid import LineOnGrid

from .utils import delete_local_srtm_files

if BASEMAP_AVAILABLE: # pragma: no cover
    from .mapping import Map




import pytest
import numpy as np
from geonum.conftest import does_not_raise_exception
from LatLon23 import LatLon, GeoVector
from geonum.geopoint import GeoPoint
from geonum.geovector3d import GeoVector3D
from geonum.exceptions import OutOfDomain
from geonum import TopoData

FAKE_TOPO = TopoData(lats=[-0.5,0.5],
                     lons=[-0.5,0.5],
                     data=[[10,10],[10,10]])

def test_GeoPoint__ALTERR_DEFAULT():
    assert GeoPoint._ALTERR_DEFAULT == 1e9

def test_GeoPoint__init__empty():
    gp = GeoPoint()
    assert isinstance(gp, LatLon)
    assert gp.name == 'undefined'
    assert gp.topo_access_mode == None
    assert gp.topo_data == None
    assert gp.local_topo_path == None
    assert gp.longitude == 0
    assert gp.latitude == 0
    assert gp.altitude == 0 # no SRTM over seas
    assert gp.altitude_err == GeoPoint._ALTERR_DEFAULT


@pytest.mark.parametrize('args,eval_args,raises', [
    ({},{}, does_not_raise_exception()),
    ({'altitude' : 10, 'auto_topo_access' : True}, {},
     pytest.raises(ValueError)),
    ({'altitude' : 10}, {'altitude' : 10},
     does_not_raise_exception()),
    ({'topo_data' : 43}, {}, pytest.raises(ValueError)),
    ({'topo_data' : FAKE_TOPO, 'auto_topo_access' : True},
     {'altitude' : 10}, does_not_raise_exception())
    ])
def test_GeoPoint__init__(args, eval_args, raises):
    with raises:
        gp = GeoPoint(**args)
        assert isinstance(gp, GeoPoint)
        for arg, val in eval_args.items():
            assert getattr(gp, arg) == val

@pytest.mark.parametrize('topo_data,gp,raises', [
    ('blaa', GeoPoint(), pytest.raises(ValueError)),
    (FAKE_TOPO, GeoPoint(), does_not_raise_exception()),
    (FAKE_TOPO, GeoPoint(10,10), pytest.raises(OutOfDomain))
    ])
def test_GeoPoint_set_topo_data(topo_data, gp, raises):
    with raises:
        gp.set_topo_data(topo_data)
        assert gp.topo_data == topo_data

@pytest.mark.parametrize('points,ll,tr', [
    ((), [44.994,14.991],[45.006,15.009]),
    ([GeoPoint(46, 16)], [44.913,14.878],[46.086,16.124]),
    ([GeoPoint(46, 16), GeoPoint(-20, -80)], [-27.464, -88.592],
     [53.016, 28.711])
])
def test_GeoPoint_range_borders(points,ll,tr):
    p0 = GeoPoint(45, 15)
    pll, ptr = p0.range_borders(*points)
    assert isinstance(pll, GeoPoint)
    assert isinstance(ptr, GeoPoint)
    _ll = [pll.latitude, pll.longitude]
    _tr = [ptr.latitude, ptr.longitude]
    np.testing.assert_allclose(_ll, ll, atol=0.01)
    np.testing.assert_allclose(_tr, tr, atol=0.01)

@pytest.mark.parametrize('kwargs,pf', [
    ({}, (45,15)),
    (dict(lat1=46, lon1=16), (46,16)),
    (dict(geo_point=GeoPoint(46,16)), (46,16)),
    (dict(azimuth=45,dist_hor=111*np.sqrt(2)),(45.99,16.433))
])
def test_GeoPoint_get_topo_data(kwargs,pf):
    p0 = GeoPoint(45, 15)
    topo, p1 = p0.get_topo_data(**kwargs)
    assert isinstance(topo, TopoData)
    assert isinstance(p1, GeoPoint)
    np.testing.assert_allclose([p1.latitude, p1.longitude], pf, atol=1e-2)


@pytest.mark.parametrize('topo,lat1,lon1,result', [
    (None,None,None,False),
    (FAKE_TOPO,None,None,True),
    (FAKE_TOPO,0.1,0.1,True),
    (FAKE_TOPO,10,20,False),
])
def test_GeoPoint_check_topo(topo,lat1,lon1,result):
    pt = GeoPoint(-0.1,-0.1,topo_data=topo)
    val = pt.check_topo(lat1,lon1)
    assert result == val

@pytest.mark.parametrize('kwargs,raises', [
    ({}, pytest.raises(ValueError)),
    (dict(geo_point=GeoPoint(46,15)), does_not_raise_exception()),
])
def test_GeoPoint_get_elevation_profile(kwargs,raises):
    from geonum.elevationprofile import ElevationProfile
    p = GeoPoint(45,15)
    with raises:
        ep = p.get_elevation_profile(**kwargs)
        assert isinstance(ep, ElevationProfile)

@pytest.mark.parametrize('vkwargs,result', [
    (dict(dx=1, dy=0), (0, -0.008983)),
    (dict(dx=111.15, dy=111.15), (-1.005, -0.999)),
    (dict(initial_heading=270, distance=222.15), (0, 1.996)),
    (dict(initial_heading=225, distance=222.15*np.sqrt(2)), (2.009, 1.996)),
])
def test_GeoPoint__sub_geo_vector_2d(vkwargs,result):
    pt = GeoPoint(latitude=0,longitude=0)
    v = GeoVector(**vkwargs)
    p2 = pt._sub_geo_vector_2d(v)
    assert isinstance(p2, GeoPoint)
    assert p2.altitude == pt.altitude
    np.testing.assert_allclose([p2.latitude, p2.longitude],
                               result, atol=0.001)

@pytest.mark.parametrize('vkwargs,result', [
    (dict(dx=1, dy=0), (0, 0.008983)),
    (dict(dx=111.15, dy=111.15), (1.005, 0.999)),
    (dict(initial_heading=270, distance=222.15), (0, -1.996)),
    (dict(initial_heading=225, distance=222.15*np.sqrt(2)), (-2.009, -1.996)),
])
def test_GeoPoint__add_geo_vector_2d(vkwargs,result):
    pt = GeoPoint(0,0)
    v = GeoVector(**vkwargs)
    p2 = pt._add_geo_vector_2d(v)
    assert isinstance(p2, GeoPoint)
    assert p2.altitude == pt.altitude
    np.testing.assert_allclose([p2.latitude, p2.longitude],
                               result, atol=0.001)

@pytest.mark.parametrize('vkwargs,result', [
    (dict(dx=1, dy=0), (0, -0.008983, 0)),
    (dict(dx=1, dy=0, dz=100), (0, -0.008983, -100)),
    (dict(dx=111.15, dy=111.15, dz=-1100), (-1.005, -0.999, 1100)),
    (dict(azimuth=270, dist_hor=222.15), (0, 1.996, 0)),
    (dict(azimuth=270, dist_hor=222.15, dz=100), (0, 1.996, -100)),
    (dict(azimuth=270, dist_hor=1, elevation=45), (0, 0.008983, -1000)),
    (dict(azimuth=225, dist_hor=np.sqrt(2), elevation=45), (0.008983, 0.008983,
                                                            -1000*np.sqrt(2))),
])
def test_GeoPoint__sub_geo_vector_3d(vkwargs,result):
    pt = GeoPoint(0,0, altitude=0)
    v = GeoVector3D(**vkwargs)
    p2 = pt._sub_geo_vector_3d(v)
    assert isinstance(p2, GeoPoint)
    np.testing.assert_allclose([p2.latitude, p2.longitude, p2.altitude],
                               result, atol=0.001)

@pytest.mark.parametrize('vkwargs,result', [
    (dict(dx=1, dy=0), (0, 0.008983, 0)),
    (dict(dx=1, dy=0, dz=100), (0, 0.008983, 100)),
    (dict(dx=111.15, dy=111.15, dz=-1100), (1.005, 0.999, -1100)),
    (dict(azimuth=270, dist_hor=222.15), (0, -1.996, 0)),
    (dict(azimuth=270, dist_hor=222.15, dz=100), (0, -1.996, 100)),
    (dict(azimuth=270, dist_hor=1, elevation=45), (0, -0.008983, 1000)),
    (dict(azimuth=225, dist_hor=np.sqrt(2), elevation=45), (-0.008983, -0.008983,
                                                            1000*np.sqrt(2))),
])
def test_GeoPoint__add_geo_vector_3d(vkwargs,result):
    pt = GeoPoint(0,0, altitude=0)
    v = GeoVector3D(**vkwargs)
    p2 = pt._add_geo_vector_3d(v)
    assert isinstance(p2, GeoPoint)
    np.testing.assert_allclose([p2.latitude, p2.longitude, p2.altitude],
                               result, atol=0.001)

def test_GeoPoint__sub_latlon():
    pt = GeoPoint(0,1, altitude=1000)
    p2 = LatLon(-1, 0)
    result = pt._sub_latlon(p2)
    assert isinstance(result, GeoVector3D)

@pytest.mark.parametrize('other,raises', [
    (42,pytest.raises(ValueError)),
    (GeoPoint(0,0),pytest.raises(ValueError)),
    (GeoVector(dx=1,dy=1), does_not_raise_exception()),
    (GeoVector3D(dx=1,dy=1, dz=100), does_not_raise_exception()),
])
def test_GeoPoint___add__(other,raises):
    pt = GeoPoint(0, 0, altitude=0)
    with raises:
        p2 = pt.__add__(other)
        assert isinstance(p2, GeoPoint)

class Fake:
    def type(self):
        return 'blaaa'

@pytest.mark.parametrize('other,tp,raises', [
    (42,None, pytest.raises(ValueError)),
    (Fake(), None, pytest.raises(ValueError)),
    (GeoVector(dx=1,dy=1), GeoPoint, does_not_raise_exception()),
    (GeoVector3D(dx=1,dy=1, dz=100), GeoPoint, does_not_raise_exception()),
    (LatLon(1,1), GeoVector3D, does_not_raise_exception()),
    (GeoPoint(1,1), GeoVector3D, does_not_raise_exception()),
])
def test_GeoPoint___sub__(other,tp,raises):
    pt = GeoPoint(0, 0, altitude=0)
    with raises:
        val = pt.__sub__(other)
        assert isinstance(val, tp)

def test_GeoPoint___str__():
    pt = GeoPoint(0, 0, altitude=0)
    st = pt.__str__()
    assert isinstance(st,str)
    assert st == 'GeoPoint undefined\nLat: 0.0,  Lon: 0.0, Alt: 0 m\n'

def test_GeoPoint___repr__():
    pt = GeoPoint(0, 0, altitude=0)
    st = pt.__repr__()
    assert isinstance(st,str)
    assert st == 'Lat: 0.0, Lon: 0.0, Alt: 0 m'

def test_GeoPoint_from_LatLon():
    pt = GeoPoint.from_LatLon(LatLon(0,1))
    assert isinstance(pt, GeoPoint)
    with pytest.raises(ValueError):
        GeoPoint.from_LatLon((10, 20))


import numpy as np
import pytest
import numpy.testing as npt
from numpy import ndarray

from geonum.conftest import does_not_raise_exception
from geonum import GeoPoint, GeoVector3D, TopoDataAccess, TopoData
from geonum import geosetup as mod
from geonum.conftest import skip_srtm
from geonum.exceptions import OutOfDomain

@pytest.mark.parametrize(
    'points,vectors,lat_ll,lon_ll,lat_tr,lon_tr,id,topo_access_mode,'
    'local_topo_path,cmap_vecs,raises',[
        (None,None,None,None,None,None,None,None,None,None,
         does_not_raise_exception()),
        (GeoPoint(),None,None,None,None,None,None,None,None,None,
         does_not_raise_exception()),
        ('Blaaaa',None,None,None,None,None,None,None,None,None,
         pytest.raises(ValueError)),
        (None,GeoVector3D(0,0),None,None,None,None,None,None,None,None,
         does_not_raise_exception()),
        (None, 'Blaaaa',None,None,None,None,None,None,None,None,
         pytest.raises(ValueError)),
        (None, None, 42, 14,43, 15, None, None, None, None,
         does_not_raise_exception()),
        (None, None, 42, None, None, None, None, None, None, None,
         does_not_raise_exception()),

    ])
def test_GeoSetup__init__(points,vectors,lat_ll,lon_ll,lat_tr,lon_tr,
                          id,topo_access_mode, local_topo_path,
                          cmap_vecs,raises):
    with raises:
        stp=mod.GeoSetup(points, vectors, lat_ll, lon_ll, lat_tr, lon_tr,
                         id, topo_access_mode, local_topo_path,
                         cmap_vecs)
        if id is None:
            id ='MyGeoSetup'
        if topo_access_mode is None:
            topo_access_mode = "srtm"
        if cmap_vecs is None:
            cmap_vecs = 'Greens'
        assert stp.id == id
        assert stp.topo_access_mode == topo_access_mode
        assert stp.local_topo_path == local_topo_path
        assert stp._cmap_vecs == cmap_vecs

@pytest.fixture(scope='module')
def empty_setup():
    return mod.GeoSetup()

@skip_srtm
def test_GeoSetup_create_test_setup():
    stp = mod.GeoSetup.create_test_setup()
    assert isinstance(stp, mod.GeoSetup)
    assert len(stp.points) == 4
    assert len(stp.vectors) == 2

@pytest.fixture(scope='module')
def dummy_setup():
    stp = mod.GeoSetup.create_test_setup()
    return stp

def test_GeoSetup_ll():
    stp = mod.GeoSetup()
    with pytest.raises(AttributeError):
        stp.ll
    stp.ll = GeoPoint(45, 15, name='ll')
    stp.set_borders_from_points()
    stp.load_topo_data()
    assert isinstance(stp.ll, GeoPoint)
    with pytest.raises(ValueError):
        stp.ll = 42
    stp.ll = GeoPoint(44,15,name='ll')

def test_GeoSetup_tr():
    stp = mod.GeoSetup()
    with pytest.raises(AttributeError):
        stp.tr
    stp.tr = GeoPoint(45, 15, name='tr')
    stp.set_borders_from_points()
    stp.load_topo_data()
    assert isinstance(stp.tr, GeoPoint)
    with pytest.raises(ValueError):
        stp.tr = 42
    stp.tr = GeoPoint(44,15,name='tr')

def test_GeoSetup_lon_ll(dummy_setup):
    npt.assert_allclose(dummy_setup.lon_ll, 14.985, atol=0.001)

def test_GeoSetup_lat_ll(dummy_setup):
    npt.assert_allclose(dummy_setup.lat_ll, 37.743, atol=0.001)

def test_GeoSetup_lon_tr(dummy_setup):
    npt.assert_allclose(dummy_setup.lon_tr, 15.025, atol=0.001)

def test_GeoSetup_lat_tr(dummy_setup):
    npt.assert_allclose(dummy_setup.lat_tr, 37.774, atol=0.001)

def test_GeoSetup_delta_lon(dummy_setup):
    npt.assert_allclose(dummy_setup.delta_lon, 0.03, atol=0.01)

def test_GeoSetup_delta_lat(dummy_setup):
    npt.assert_allclose(dummy_setup.delta_lat, 0.03, atol=0.01)

def test_GeoSetup_borders_set():
    assert mod.GeoSetup().borders_set == False
    assert mod.GeoSetup(points=GeoPoint(0,0)).borders_set == True

def test_GeoSetup_center_coordinates():
    gs = mod.GeoSetup()
    with pytest.raises(AttributeError):
        coords = gs.center_coordinates
    gs.add_geo_point(GeoPoint(-0.5,-0.5,name='ll'))
    with pytest.raises(AttributeError):
        coords = gs.center_coordinates
    gs.add_geo_point(GeoPoint(0.5, 0.5, name='tr'))
    coords = gs.center_coordinates
    assert coords == (0,0)

def test_GeoSetup_magnitude(dummy_setup):
    npt.assert_allclose(dummy_setup.magnitude, 4.898, atol=0.01)

def test_GeoSetup_topo_access(empty_setup):
    assert isinstance(empty_setup.topo_access, TopoDataAccess)

def test_GeoSetup_cmap_vecs(empty_setup):
    assert empty_setup._cmap_vecs == 'Greens'
    from matplotlib.colors import LinearSegmentedColormap
    assert isinstance(empty_setup.cmap_vecs, LinearSegmentedColormap)

def test_GeoSetup_has_points(empty_setup):
    assert not empty_setup.has_points()
    empty_setup.add_geo_point(GeoPoint(name='null'), assert_in_domain=False)
    assert empty_setup.has_points()
    empty_setup.delete_geo_point('null')
    assert not empty_setup.has_points()

@skip_srtm
def test_GeoSetup_has_topo_data(dummy_setup):
    assert dummy_setup.has_topo_data() == False
    dummy_setup.get_topo()
    assert dummy_setup.has_topo_data() == True
    dummy_setup.topo_data = None

def test_GeoSetup_set_local_topo_path(tmpdir):
    gs = mod.GeoSetup()
    fp = str(tmpdir)
    gs.set_local_topo_path(tmpdir)
    assert gs.local_topo_path == fp
    with pytest.raises(FileExistsError):
        gs.set_local_topo_path('/blaaaaaa/blub')

@pytest.mark.parametrize('new_mode,local_path,raises', [
    ('bla', None, does_not_raise_exception()),
    ('bla', 'blub', pytest.raises(FileExistsError)),
    ('bla', '/', does_not_raise_exception()),
    (None, None, does_not_raise_exception()),

])
def test_GeoSetup_change_topo_mode(new_mode,local_path,raises):
    gs = mod.GeoSetup()
    with raises:
        gs.change_topo_mode(new_mode,local_path)

@skip_srtm
def test_GeoSetup_get_topo():
    stp = mod.GeoSetup.create_test_setup()
    stp.get_topo()
    assert isinstance(stp.topo_data, TopoData)

@skip_srtm
def test_GeoSetup_load_topo_data():
    stp = mod.GeoSetup()
    stp.add_geo_point(GeoPoint(45,15))
    stp.load_topo_data()
    assert isinstance(stp.topo_data, TopoData)

def test_GeoSetup_has_point():
    stp = mod.GeoSetup(lat_ll=45, lon_ll=15, lat_tr=50, lon_tr=20)
    assert stp.has_point('ll') == True
    assert stp.has_point('bla') == False
    stp.add_geo_point(GeoPoint(10, 20, name='bla'), assert_in_domain=False)
    assert stp.has_point('bla') == True

def test_GeoSetup_has_vector():
    stp = mod.GeoSetup(lat_ll=45, lon_ll=15, lat_tr=50, lon_tr=20)
    assert stp.has_vector('bla') == False
    stp.add_geo_vector(GeoVector3D(1,1,1,name='bla'))
    assert stp.has_vector('bla') == True

def test_GeoSetup_contains_coordinate():
    stp = mod.GeoSetup(lat_ll=45, lon_ll=15, lat_tr=50, lon_tr=20)
    assert stp.contains_coordinate(46, 16) == True
    assert stp.contains_coordinate(30, 16) == False

@pytest.mark.parametrize('pt,assert_in_domain,overwrite_existing,raises', [
    (42,False,True,pytest.raises(ValueError)),
    (GeoPoint(0,0,name='bla'),False,True,does_not_raise_exception()),
    (GeoPoint(0,0,name='bla'),False,False,pytest.raises(ValueError)),
    (GeoPoint(0,0,name='bla'),True,True,pytest.raises(OutOfDomain)),
    (GeoPoint(0,0,name='ll'),True,False,does_not_raise_exception()),
    (GeoPoint(0,0,name='tr'),True,False,does_not_raise_exception()),
])
def test_GeoSetup_add_geo_point(pt,assert_in_domain,overwrite_existing,raises):
    stp = mod.GeoSetup(lat_ll=45, lon_ll=15, lat_tr=50, lon_tr=20)
    stp.points['bla'] = GeoPoint(47,16,name='bla')
    stp.topo_data = TopoData(lats=[45,50], lons=[15,20], data=np.ones((2,2)))
    with raises:
        stp.add_geo_point(pt=pt,assert_in_domain=assert_in_domain,
                          overwrite_existing=overwrite_existing)
        assert pt.name in stp.points


@pytest.mark.parametrize('pts,assert_in_domain,raises', [
    ([GeoPoint(0,0)], True, pytest.raises(OutOfDomain)),
    ([GeoPoint(48,16)], True, does_not_raise_exception()),
    ([GeoPoint(48,16), GeoPoint(49,16)], True, pytest.raises(ValueError)),
    ([GeoPoint(48,16,name='p1'), GeoPoint(49,16,name='p2')], True,
    does_not_raise_exception()),
    ([GeoPoint(48,16,name='p1'), GeoPoint(49,21)], True, pytest.raises(
        OutOfDomain)),
    ([GeoPoint(48,16,name='p1'), GeoPoint(49,21)], False, does_not_raise_exception()),
])
def test_GeoSetup_add_geo_points(pts,assert_in_domain,raises):
    stp = mod.GeoSetup(lat_ll=45,lon_ll=15, lat_tr=50, lon_tr=20)
    with raises:
        stp.add_geo_points(*pts, assert_in_domain=assert_in_domain)
        assert len(stp.points) == len(pts) + 2


@pytest.mark.parametrize('vec,overwrite_existing,raises', [
    (42, False, pytest.raises(ValueError)),
    (GeoVector3D(1,1,1,name='bla'), False, pytest.raises(ValueError)),
    (GeoVector3D(1,1,1,name='bla'), True, does_not_raise_exception()),
    (GeoVector3D(1,1,1,name='blub'), False, does_not_raise_exception()),
])
def test_GeoSetup_add_geo_vector(vec,overwrite_existing,raises):
    stp = mod.GeoSetup(lat_ll=45, lon_ll=15, lat_tr=50, lon_tr=20)
    stp.vectors['bla'] = GeoVector3D(1,1,1,name='bla')
    with raises:
        stp.add_geo_vector(vec=vec,
                          overwrite_existing=overwrite_existing)
        assert vec.name in stp.vectors


@pytest.mark.parametrize('vecs,raises', [
    ([GeoVector3D(1,1,1,name='bla')], does_not_raise_exception()),
    ([GeoVector3D(1,1,1,name='bla'),
      GeoVector3D(1,1,1,name='bla')], does_not_raise_exception()),
])
def test_GeoSetup_add_geo_vectors(vecs,raises):
    stp = mod.GeoSetup(lat_ll=45, lon_ll=15, lat_tr=50, lon_tr=20)
    with raises:
        stp.add_geo_vectors(*vecs)

def test_GeoSetup_delete_geo_point():
    stp = mod.GeoSetup(lat_ll=45, lon_ll=15, lat_tr=50, lon_tr=20)
    with pytest.raises(ValueError):
        stp.delete_geo_point('bla')
    stp.add_geo_point(GeoPoint(0,0,name='bla'),assert_in_domain=False)
    assert 'bla' in stp.points
    stp.delete_geo_point('bla')
    assert not 'bla' in stp.points

def test_GeoSetup_delete_geo_vector():
    stp = mod.GeoSetup(lat_ll=45, lon_ll=15, lat_tr=50, lon_tr=20)
    with pytest.raises(ValueError):
        stp.delete_geo_vector('bla')
    stp.add_geo_vector(GeoVector3D(1,1,1,name='bla'))
    assert 'bla' in stp.vectors
    stp.delete_geo_vector('bla')
    assert not 'bla' in stp.vectors

def test_GeoSetup_new_geo_point():
    stp = mod.GeoSetup(lat_ll=45, lon_ll=15, lat_tr=50, lon_tr=20)
    with pytest.raises(OutOfDomain):
        stp.new_geo_point(latitude=10, longitude=20)

    stp.new_geo_point(latitude=46, longitude=19, name='name')
    stp.new_geo_point(47, 19, name='name2')

def test_GeoSetup__all_lats_lons():
    stp = mod.GeoSetup(lat_ll=45, lon_ll=15, lat_tr=50, lon_tr=20)
    result = stp._all_lats_lons()
    assert isinstance(result[0], ndarray)
    assert isinstance(result[1], ndarray)
    assert list(result[0]) == [45, 50]
    assert list(result[1]) == [15, 20]

@pytest.mark.parametrize('pts,extend_km,to_square,raises,delta_lat,'
                         'delta_lon', [
    ([],1,True,pytest.raises(AttributeError),None,None),
    ([GeoPoint(0,0)],1,True,does_not_raise_exception(),0.013,0.013),
    ([GeoPoint(0,0)],111.15/np.sqrt(2),True,does_not_raise_exception(),
     1.005,0.999),
    ([GeoPoint(0, 0, name='p1'), GeoPoint(1, 10, name='p2')], 0, True,
     does_not_raise_exception(),
     10.061, 10),
    ([GeoPoint(0, 0, name='p1'), GeoPoint(1, 10, name='p2')], 0, False,
     does_not_raise_exception(),
     1, 10),
    ([GeoPoint(0, 0, name='p1'), GeoPoint(1, 10, name='p2')], 100,
     False,
     does_not_raise_exception(),
     2.279, 11.271),
])
def test_set_borders_from_points(pts,extend_km,to_square,raises,delta_lat,
                                 delta_lon):
    stp = mod.GeoSetup(points=pts, init_borders=False)
    with raises:
        stp.set_borders_from_points(extend_km,to_square)
        npt.assert_allclose(delta_lat, stp.delta_lat, atol=0.001)
        npt.assert_allclose(delta_lon, stp.delta_lon, atol=0.001)



near_Etna = GeoPoint(latitude=37.751, longitude=14.993,
                     name="near Etna", auto_topo_access=True)

@skip_srtm
@pytest.mark.parametrize('p,radius,result,raises', [
    (None,None,[],pytest.raises(ValueError)),
    (near_Etna,None,['Etna'],does_not_raise_exception()),
    (near_Etna,0.01,[],does_not_raise_exception()),
    (near_Etna,20,['Etna', 'Observatory', 'll', 'tr'],does_not_raise_exception()),
])
def test_GeoSetup_points_close(p,radius,result,raises):
    stp = mod.GeoSetup.create_test_setup()
    with raises:
        val = stp.points_close(p,radius)
        assert val == result





# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

class InvalidTopoMode(IOError):
    pass

class IntersectNotFound(ValueError):
    pass

class OutOfDomain(ValueError):
    pass

class SRTMNotCoveredError(ValueError):
    pass

class TopoAccessError(IOError):
    pass